"""
Package Purpose: tools related to xarrays
"""
from . import _xarray_bind_tools  # allow _xarray_bind_tools to bind stuff, as needed

from .xarray_accessors import (
    xarray_register_dataarray_accessor_cls,
    xarray_register_dataset_accessor_cls,
    XarrayAccessor,
    pcAccessor, pcArrayAccessor, pcDatasetAccessor,
)
from .xarray_agg_stats import (
    xarray_aggregate,
    xarray_sum,
    # stats
    xarray_stats,
    xarray_min, xarray_mean, xarray_median, xarray_max,
    xarray_std, xarray_rms,
    # non-aggregator stats
    xarray_minimum, xarray_minimum_of_datavars,
    xarray_maximum, xarray_maximum_of_datavars,
)
from .xarray_coords import (
    # coords
    xarray_nondim_coords, xarray_dims_coords,
    xarray_assign_self_as_coord,
    xarray_fill_coords, xarray_index_coords,
    xarray_promote_index_coords, xarray_demote_index_coords,
    xarray_scale_coords, xarray_shift_coords, xarray_log_coords,
    xarray_is_sorted,
    # coord math
    xarray_get_dx_along, xarray_differentiate,
)
from .xarray_dimensions import (
    _paramdocs_ensure_dims,
    # dimensions
    is_iterable_dim,
    take_along_dimension, take_along_dimensions, join_along_dimension,
    xarray_rename, xarray_assign, xarray_promote_dim, xarray_ensure_dims,
    xarray_squeeze, xarray_squeeze_close, xarray_closeness,
    xarray_drop_unused_dims, xarray_drop_vars, xarray_popped,
    # broadcasting
    xarray_broadcastable_array, xarray_broadcastable_from_dataset,
    xarray_from_broadcastable, 
    # coarsen / windowing
    xarray_coarsened,
)
from .xarray_grids import (
    xr1d, xrrange, xarray_range,
    XarrayGrid, xarray_grid,
    xarray_angle_grid,
)
from .xarray_indexing import (
    xarray_isel, xarray_search, xarray_sel,
    xarray_where,
    xarray_map,
    xarray_argsort, xarray_sort_array_along, xarray_sort_dataset_along,
    xarray_cmin, xarray_cmax, xarray_varmin, xarray_varmax,
    xarray_at_min_of, xarray_at_max_of,
    xarray_min_coord_where, xarray_max_coord_where,
)
from .xarray_io import (
    XarrayIoSerializable,
    xarray_save, xarray_load, xarray_mergeload,
    _xarray_save_prep,
    _xarray_coord_serializations, _xarray_coord_deserializations,
)
from .xarray_masks import (
    xarray_mask,
    xarray_has_mask, xarray_store_mask, xarray_stored_mask, xarray_popped_mask,
    xarray_unmask,
    xarray_demask_from_mask, xarray_demask_from_ds,
    xarray_unmask_var, xarray_unmask_vars,
)
from .xarray_misc import (
    xarray_copy_kw, xarray_with_data,
    xarray_as_array,
    xarray_vars_lookup, xarray_vars_lookup_with_defaults,
    xarray_where_finite,
    xarray_astypes, xarray_convert_types,
    xarray_object_coords_to_str,
)
from .xarray_sci import (
    xarray_interp_inverse,
    xarray_gaussian_filter,
    xarray_polyfit, xarray_assign_polyfit_stddev, xarray_coarsened_polyfit,
    xarray_curve_fit, xarray_curve_eval, XarrayCurveFitter,
    XarrayLineFitter, xarray_line_fit,
)
from .xarray_werr_stats import (
    xarray_werrmean,
    xarray_werr2pmstd, xarray_pmstd2werr, xarray_mean_pm_std,
    xarray_werradd, xarray_werrsub, xarray_werrmul, xarray_werrdiv,
)



"""
File Purpose: base & simple derived quantities.

The way to load BASE_QUANTS should be defined in hookups,
    as they will depend on the kind of input data being loaded.

The way to load SIMPLE_DERIVED_QUANTS is defined in this file,
    as they depend only on BASE_QUANTS.
    However, they are decently likely to be overwritten, depending on kind of input.
        E.g. Ebysus saves 'r' and 'm', not 'n'.
            Since 'n' is a base quant, EbysusCalculator needs to overwrite it anyways,
                however it will use something like n = r / m, rather than n = number density from file.
            Meanwhile, EbysusCalculator reads 'r' from file, rather than r = n * m.

Other parts of PlasmaCalcs may assume it is possible to load BASE_QUANTS and SIMPLE_DERIVED_QUANTS.
    PlasmaCalcs will fail with LoadingNotImplementedError when the way to load the quant
    hasn't been defined yet, for the kind of input data being loaded.
"""
import numpy as np

from .quantity_loader import QuantityLoader
from ..errors import LoadingNotImplementedError

# -- this dict is purely for documentation purposes, and not used by the code --
BASE_QUANTS = {
    # -- non-plasma quantities --
    'ds': 'vector(spatial scale), e.g. [dx, dy, dz]',
    # -- fluid constant quantities --
    'm': 'mass',    # of a "single particle". for protons, ~= +1 atomic mass unit
    'q': 'charge',  # of a "single particle". for protons, == +1 elementary charge
    'gamma': 'adiabatic index',
    # -- fluid quantities --
    'n': 'number density',
    'u': 'velocity',    # vector quantity (depends on self.component)
    'T': 'temperature', # "maxwellian" temperature (classical T in thermodynamics)
    'nusj': 'collision frequency',  # for a single particle of s to collide with any of j
    # -- global electromagnetic quantities --
    'E': 'electric field',
    'B': 'magnetic field',
    # -- neutral quantities --
    'm_neutral': 'mass of neutral(s)',    # of a "single particle". for Hydrogen, ~= +1 atomic mass unit
    'n_neutral': 'number density of neutral(s)',
    'u_neutral': 'velocity of neutral(s)',   # vector quantity.
    'T_neutral': 'temperature of neutral(s)',  # "maxwellian" temperature (classical T in thermodynamics)
}

# -- this dict is purely for documentation purposes, and not used by the code --
SIMPLE_DERIVED_QUANTS = {
    # -- non-E&M quantities --
    'r': 'mass density', 
    'p': 'momentum density',  # note - lowercase
    'P': 'pressure ("isotropic/maxwellian")',      # note - uppercase
    'Tjoule': 'temperature ("isotropic/maxwellian"), in energy units (multipled by kB). Joules if SI units',
    'e': 'energy density',
    'nusn': 'collision frequency with neutrals',  # for a single particle of s to collide with any neutral
    'nuns': 'collision frequency of neutrals with s',  # for a single neutral to collide with any s
    # -- E&M quantities --
    'nq': 'charge density',
    'Jf': 'current density (associated with fluid)',  # per area, e.g. A/m^2
    'J': 'total current density',  # per area, e.g. A/m^2
    'E_un0': 'electric field in the u_neutral=0 frame',
}

class AllBasesLoader(QuantityLoader):
    '''all base quantities.
    The implementation here just raises LoadingNotImplementedError, for all the BASE_QUANTS.

    Subclasses should override these methods to load the quantities as appropriate,
        probably either from a file or from calculations involving other quantities loaded from files.

    See also: BASE_QUANTS
    '''
    # # # NON-PLASMA QUANTITIES # # #
    @known_var
    def get_ds(self):
        '''vector(spatial scale), e.g. [dx, dy, dz]
        [Not implemented for this class]
        '''
        raise LoadingNotImplementedError('ds')

    # # # FLUID CONSTANT QUANTITIES # # #
    @known_var
    def get_m(self):
        '''mass, of a "single particle". For protons, ~= +1 atomic mass unit.
        [Not implemented for this class]
        '''
        raise LoadingNotImplementedError('m')

    @known_var
    def get_q(self):
        '''charge, of a "single particle". for protons, == +1 elementary charge.
        [Not implemented for this class]
        '''
        raise LoadingNotImplementedError('q')

    @known_var
    def get_gamma(self):
        '''adiabatic index.
        [Not implemented for this class]
        '''
        raise LoadingNotImplementedError('gamma')

    # # # FLUID QUANTITIES # # #
    @known_var
    def get_n(self):
        '''number density.
        [Not implemented for this class]
        '''
        raise LoadingNotImplementedError('n')

    @known_var(dims=['component'])
    def get_u(self):
        '''velocity. vector quantity (result depends on self.component)
        [Not implemented for this class]
        '''
        raise LoadingNotImplementedError('u')

    @known_var
    def get_T(self):
        '''temperature. "maxwellian" temperature (classical T in thermodynamics).
        [Not implemented for this class]
        '''
        raise LoadingNotImplementedError('T')

    @known_var
    def get_nusj(self):
        '''collision frequency. for a single particle of s to collide with any of j.
        [Not implemented for this class]
        '''
        raise LoadingNotImplementedError('nusj')

    # # # GLOBAL ELECTROMAGNETIC QUANTITIES # # #
    @known_var
    def get_E(self):
        '''electric field.
        [Not implemented for this class]
        '''
        raise LoadingNotImplementedError('E')

    @known_var
    def get_B(self):
        '''magnetic field.
        [Not implemented for this class]
        '''
        raise LoadingNotImplementedError('B')

    # # # NEUTRAL QUANTITIES # # #
    # (since it's common for plasma codes to treat neutrals in some special way.)
    @known_var(aliases=['m_n'], ignores_dims=['fluid'])
    def get_m_neutral(self):
        '''mass, of a "single neutral particle". For Hydrogen, ~= +1 atomic mass unit.
        [Uses self.get_neutral('m') if possible, else crash. Subclass may override.]
        '''
        if hasattr(self, 'get_neutral'):  # (works if self is a FluidHaver with neutral fluids.)
            return self.get_neutral('m')
        # else, crash. provided 'm_neutral' var separately because sometimes non-FluidHavers
        #   (e.g. single-fluid mhd) might still have ideas about "what is the neutral fluid doing?".
        raise LoadingNotImplementedError('m_neutral')

    @known_var(aliases=['n_n'], ignores_dims=['fluid'])
    def get_n_neutral(self):
        '''number density of neutrals.
        [Uses self.get_neutral('n') if possible, else crash. Subclass may override.]
        '''
        if hasattr(self, 'get_neutral'):
            return self.get_neutral('n')
        raise LoadingNotImplementedError('n_neutral')

    @known_var(aliases=['u_n'], dims=['component'], ignores_dims=['fluid'])
    def get_u_neutral(self):
        '''velocity of neutrals. vector quantity (result depends on self.component)
        [Uses self.get_neutral('u') if possible, else crash. Subclass may override.]
        '''
        if hasattr(self, 'get_neutral'):
            return self.get_neutral('u')
        raise LoadingNotImplementedError('u_neutral')

    @known_var(aliases=['T_n'], ignores_dims=['fluid'])
    def get_T_neutral(self):
        '''temperature of neutrals. "maxwellian" temperature (classical T in thermodynamics).
        [Uses self.get_neutral('T') if possible, else crash. Subclass may override.]
        '''
        if hasattr(self, 'get_neutral'):
            return self.get_neutral('T')
        raise LoadingNotImplementedError('T_neutral')


class SimpleDerivedLoader(QuantityLoader):
    '''simple quantities derived from the base quantities.
    Subclasses are decently-likely to override these methods, depending on the kind of input,
        because these provide similar information to the base quantities so different kinds of input
        might save these instead of saving base quantities.
        (E.g. Ebysus saves 'r' instead of 'n', so EbysusCalculator
            will override get_r to read from file, and get_n to n = r / m.)

    See also: SIMPLE_DERIVED_QUANTS
    '''
    # # # NON-E&M QUANTITIES # # #
    @known_var(deps=['n', 'm'])
    def get_r(self):
        '''mass density. r = (n * m) = (number density * mass)'''
        return self('n') * self('m')

    @known_var(deps=['u', 'r'])
    def get_p(self):
        '''momentum density. p = (u * r) = (velocity * mass density).'''
        return self('u') * self('r')

    @known_var(deps=['n', 'Tjoule'])
    def get_P(self):
        '''pressure ("isotropic/maxwellian"). P = (n * Tjoule) = (number density * T [energy units])'''
        return self('n') * self('Tjoule')

    @known_var(deps=['T'])
    def get_Tjoule(self):
        '''temperature ("isotropic/maxwellian"), in energy units. Tjoule = kB * T.
        If using SI units, result will be in Joules.
        '''
        return self.u('kB') * self('T')

    @known_var(deps=['gamma', 'P'])
    def get_e(self):
        '''energy density. e = P / (gamma - 1) = pressure / (adiabatic index - 1)'''
        return self('P') / (self('gamma') - 1)

    @known_var(deps=['nusj'], ignores_dims=['jfluid'])
    def get_nusn(self):
        '''collision frequency. for a single particle of s to collide with any neutral.
        Computed as self('nusj', jfluid=self.jfluids.get_neutral()).
        '''
        neutral = self.jfluids.get_neutral()  # if crash, subclass should implement nusn separately.
        with self.using(jfluid=neutral):
            return self('nusj')

    @known_var(deps=['nusn', 'm/m_n', 'n/n_n'])
    def get_nuns(self):
        '''collision frequency. for a single neutral particle to collide with any s.
        nuns = nusn * (m / m_neutral) * (n / n_neutral).
        (from conservation of momentum, and summing collisional momentum transfer between species)
        '''
        return self('nusn') * self('m/m_n') * self('n/n_n')

    # # # E&M QUANTITIES # # #
    @known_var(deps=['n', 'q'])
    def get_nq(self):
        '''charge density. nq = (n * q) = (number density * charge)'''
        return self('n') * self('q')

    @known_var(deps=['nq', 'u'])
    def get_Jf(self):
        '''current density (associated with fluid). Jf = (nq * u) = (charge density * velocity)
        This is per unit area, e.g. the SI units would be Amperes / meter^2.

        (If self is not a FluidHaver, this will equal the total current density.)
        '''
        return self('nq') * self('u')

    @known_var(deps=['Jf'], ignores_dims=['fluid'])
    def get_J(self):
        '''total current density. J = sum_across_fluids(n*q*u)
        This is per unit area, e.g. the SI units would be Amperes / meter^2.
        '''
        try:
            fluids = self.fluids
        except AttributeError:
            errmsg = f'J, for object of type {type(self)} which has no .fluids attribute'
            raise LoadingNotImplementedError(errmsg) from None
        charged = fluids.charged()  # [EFF] exclude uncharged fluids, for efficiency.
        Jfs = self('Jf', fluid=charged)
        return Jfs.sum('fluid')  # [TODO] handle "only 1 fluid" case?

    @known_var(deps=['E', 'u_n'])
    def get_E_un0(self):
        '''electric field in the u_neutral=0 frame.
        Here, asserts all of self('u_n')==0, then returns self('E').
        if the assertion fails, raise NotImplementedError (expect subclass to handle it).
        '''
        if not np.all(self('u_n', component=None)==0):
            raise NotImplementedError('E_un0 implementation here assumes u_n=0')
        return self('E')


class DirectBasesLoader(AllBasesLoader, SimpleDerivedLoader):
    '''AllBasesLoader & SimpleDerivedLoader, but load_direct when available.'''
    # # # BASES # # #
    @known_var
    def get_ds(self):
        '''vector(spatial scale), e.g. [dx, dy, dz].
        The implementation here just does self.load_direct('ds').
        '''
        return self.load_direct('ds')

    @known_var
    def get_m(self):
        '''mass, of a "single particle". For protons, ~= +1 atomic mass unit.
        The implementation here just does self.load_direct('m').
        '''
        return self.load_direct('m')

    @known_var
    def get_q(self):
        '''charge, of a "single particle". for protons, == +1 elementary charge.
        The implementation here just does self.load_direct('q').
        '''
        return self.load_direct('q')

    @known_var
    def get_gamma(self):
        '''adiabatic index.
        The implementation here just does self.load_direct('gamma').
        '''
        return self.load_direct('gamma')

    @known_var
    def get_n(self):
        '''number density.
        The implementation here just does self.load_direct('n').
        '''
        return self.load_direct('n')

    @known_var
    def get_u(self):
        '''velocity (or speed, if self doesn't have vector component dimension)
        The implementation here just does self.load_direct('u').
        '''
        return self.load_direct('u')

    @known_var
    def get_T(self):
        '''temperature. "maxwellian" temperature (classical T in thermodynamics).
        The implementation here just does self.load_direct('T').
        '''
        return self.load_direct('T')

    @known_var
    def get_nusj(self):
        '''collision frequency. for a single particle of s to collide with any of j.
        The implementation here just does self.load_direct('nusj').
        '''
        return self.load_direct('nusj')

    @known_var
    def get_E(self):
        '''electric field.
        The implementation here just does self.load_direct('E').
        '''
        return self.load_direct('E')

    @known_var
    def get_B(self):
        '''magnetic field.
        The implementation here just does self.load_direct('B').
        '''
        return self.load_direct('B')

    @known_var(aliases=['m_n'])
    def get_m_neutral(self):
        '''mass, of a "single neutral particle". For Hydrogen, ~= +1 atomic mass unit.
        self.load_direct('m_neutral') if possible, else self.load_direct('m_n')
        '''
        if 'm_neutral' in self.directly_loadable_vars():
            return self.load_direct('m_neutral')
        else:
            return self.load_direct('m_n')

    @known_var(aliases=['n_n'])
    def get_n_neutral(self):
        '''number density of neutrals.
        self.load_direct('n_neutral') if possible, else self.load_direct('n_n')
        '''
        if 'n_neutral' in self.directly_loadable_vars():
            return self.load_direct('n_neutral')
        else:
            return self.load_direct('n_n')

    @known_var(aliases=['u_n'])
    def get_u_neutral(self):
        '''velocity (or speed) of neutrals.
        self.load_direct('u_neutral') if possible, else self.load_direct('u_n')
        '''
        if 'u_neutral' in self.directly_loadable_vars():
            return self.load_direct('u_neutral')
        else:
            return self.load_direct('u_n')

    @known_var(aliases=['T_n'])
    def get_T_neutral(self):
        '''temperature of neutrals. "maxwellian" temperature (classical T in thermodynamics).
        self.load_direct('T_neutral') if possible, else self.load_direct('T_n')
        '''
        if 'T_neutral' in self.directly_loadable_vars():
            return self.load_direct('T_neutral')
        else:
            return self.load_direct('T_n')

    # # # SIMPLE DERIVED # # #
    can_load_direct_r = property(lambda self: 'r' in self.directly_loadable_vars())
    @known_var(attr_deps=[('can_load_direct_r', {True: [], False: ['n', 'm']})])
    def get_r(self):
        '''mass density. self.load_direct('r') if possible, else r = (n * m)'''
        if self.can_load_direct_r:
            return self.load_direct('r')
        return super().get_r()

    can_load_direct_p = property(lambda self: 'p' in self.directly_loadable_vars())
    @known_var(attr_deps=[('can_load_direct_p', {True: [], False: ['u', 'r']})])
    def get_p(self):
        '''momentum density. self.load_direct('p') if possible, else p = (u * r)'''
        if self.can_load_direct_p:
            return self.load_direct('p')
        return super().get_p()

    can_load_direct_P = property(lambda self: 'P' in self.directly_loadable_vars())
    @known_var(attr_deps=[('can_load_direct_P', {True: [], False: ['n', 'Tjoule']})])
    def get_P(self):
        '''pressure ("isotropic/maxwellian"). self.load_direct('P') if possible, else P = (n * Tjoule)'''
        if self.can_load_direct_P:
            return self.load_direct('P')
        return super().get_P()

    can_load_direct_Tjoule = property(lambda self: 'Tjoule' in self.directly_loadable_vars())
    @known_var(attr_deps=[('can_load_direct_Tjoule', {True: [], False: ['T']})])
    def get_Tjoule(self):
        '''temperature ("isotropic/maxwellian"), in energy units.
        self.load_direct('Tjoule') if possible, else Tjoule = kB * T
        '''
        if self.can_load_direct_Tjoule:
            return self.load_direct('Tjoule')
        return super().get_Tjoule()

    can_load_direct_e = property(lambda self: 'e' in self.directly_loadable_vars())
    @known_var(attr_deps=[('can_load_direct_e', {True: [], False: ['P', 'gamma']})])
    def get_e(self):
        '''energy density. self.load_direct('e') if possible, else e = P / (gamma - 1)'''
        if self.can_load_direct_e:
            return self.load_direct('e')
        return super().get_e()

    can_load_direct_nusn = property(lambda self: 'nusn' in self.directly_loadable_vars())
    @known_var(attr_deps=[('can_load_direct_nusn', {True: [], False: ['nusj']})])
    def get_nusn(self):
        '''collision frequency. for a single particle of s to collide with any neutral.
        self.load_direct('nusn') if possible, else nusn = nusj(neutral)
        '''
        if self.can_load_direct_nusn:
            return self.load_direct('nusn')
        return super().get_nusn()

    can_load_direct_nq = property(lambda self: 'nq' in self.directly_loadable_vars())
    @known_var(attr_deps=[('can_load_direct_nq', {True: [], False: ['n', 'q']})])
    def get_nq(self):
        '''charge density. self.load_direct('nq') if possible, else nq = (n * q)'''
        if self.can_load_direct_nq:
            return self.load_direct('nq')
        return super().get_nq()

    can_load_direct_Jf = property(lambda self: 'Jf' in self.directly_loadable_vars())
    @known_var(attr_deps=[('can_load_direct_Jf', {True: [], False: ['nq', 'u']})])
    def get_Jf(self):
        '''current density (associated with fluid). self.load_direct('Jf') if possible, else Jf = (nq * u)'''
        if self.can_load_direct_Jf:
            return self.load_direct('Jf')
        return super().get_Jf()

    can_load_direct_J = property(lambda self: 'J' in self.directly_loadable_vars())
    @known_var(attr_deps=[('can_load_direct_J', {True: [], False: ['Jf']})])
    def get_J(self):
        '''total current density. self.load_direct('J') if possible, else J = sum_across_fluids(n*q*u)'''
        if self.can_load_direct_J:
            return self.load_direct('J')
        return super().get_J()

    can_load_direct_E_un0 = property(lambda self: 'E_un0' in self.directly_loadable_vars())
    @known_var(attr_deps=[('can_load_direct_E_un0', {True: [], False: ['E', 'u_n']})])
    def get_E_un0(self):
        '''electric field in the u_neutral=0 frame.
        self.load_direct('E_un0') if possible, else E_un0 = E
        '''
        if self.can_load_direct_E_un0:
            return self.load_direct('E_un0')
        return super().get_E_un0()



"""
File Purpose: loading Bifrost number densities
"""
import numpy as np
import xarray as xr

from ....defaults import DEFAULTS
from ....errors import FormulaMissingError
from ....mhd import MhdMultifluidDensityLoader, Specie
from ....tools import simple_property

class BifrostMultifluidDensityLoader(MhdMultifluidDensityLoader):
    '''density quantities based on Bifrost single-fluid values, & inferred multifluid properties.'''

    cls_behavior_attrs.register('did_hion', default=False)
    did_hion = simple_property('_did_hion', setdefaultvia='_default_did_hion',
        doc='''whether this BifrostMultifluidCalculator has params['do_hion'] == True.
        Default is to just use the value from params.get('do_hion', False).
        when determining whether neq values exist (see self.N_MODE_OPTIONS), checks self.did_hion.

        User may wish to explicitly set did_hion, e.g. did_hion=
            =False to disable neq if bifrost had do_hion=True but hion file is missing or corrupted;
            =True to enable neq if bifrost had do_hion=False but hion file is provided separately...?''')
    def _default_did_hion(self):
        '''return default value for self.did_hion. Equivalent: self.params.get('do_hion', False).'''
        return self.params.get('do_hion', False)

    # # # N_MODE DISPATCH / CODE ARCHITECTURE # # #
    N_MODE_OPTIONS = {**MhdMultifluidDensityLoader.N_MODE_OPTIONS,
        'best': '''use best mode available, based on fluid:
            electron --> 'neq' if simulation neq enabled, else 'table'.
            H or He Specie --> 'neq' if simulation neq enabled, else 'saha'
            other Specie --> 'saha'.''',
        'neq': '''load value directly from file if simulation neq enabled and relevant to fluid.
            (crash if not electron)''',
        'QN_neq': '''sum of qi ni across self.fluids; getting 'ne' for saha via 'neq' method.
            (crash if not electron)''',
    }

    NE_MODE_OPTIONS = {**MhdMultifluidDensityLoader.NE_MODE_OPTIONS,
        'best': '''use 'neq' if simulation neq enabled, else 'table'.''',
        'neq': '''load value directly from file if simulation neq enabled.
            neq possibly available in aux:
                e-     --> 'hionne'
                H_I    --> sum('n1', 'n2', 'n3', 'n4', 'n5')
                H_II   --> 'n6'
                He_I   --> 'nhe1'  (actually, exp('nhe1'); aux stores log values)
                He_II  --> 'nhe2'  (actually, exp('nhe2'); aux stores log values)
                He_III --> 'nhe3'  (actually, exp('nhe3'); aux stores log values)
            (if not possible, crash or return NaN, depending on self.typevar_crash_if_nan)''',
        'QN_neq': '''sum of qi ni across self.fluids; getting 'ne' for saha via 'neq' method.''',
    }

    _NE_MODE_INTERNAL = {**MhdMultifluidDensityLoader._NE_MODE_INTERNAL,
        'QN_neq': 'neq'
    } 

    # # # GENERIC NUMBER DENSITY # # #
    @known_var(load_across_dims=['fluid'])
    def get_ntype(self):
        '''ntype of self.fluid. Result depends on fluid as well as self.n_mode (and ne_mode if electron).
        See self.N_MODE_OPTIONS and self.NE_MODE_OPTIONS for options.
        possible results (possible in parent class)
            'SINGLE_FLUID' <--> n for SINGLE_FLUID
            'elem' <--> n for Element
            'saha' <--> n from saha equation. (not available for electrons)
            'table' <--> n from EOS table. (only available for electrons)
            'QN_table' <--> n from sum of qi ni across self.fluids, with ne from table.
            'nan' <--> nan or crash, depending on self.typevar_crash_if_nan.
        possible results (possible here but not in parent class)
            'neq' <--> n from neq-related files.
            'QN_neq' <--> n from sum of qi ni across self.fluids, with ne from neq.
        '''
        result = 'nan'  # <-- nan unless set to something else below
        f = self.fluid
        if not isinstance(f, Specie):
            result = super().get_ntype()
        else:  # f is a Specie. fully handled here for readability, despite repetitiveness with super().
            result = 'nan'
            # bookkeeping:
            e = f.is_electron()
            mode = self.ne_mode_explicit if e else self.n_mode
            H = (f.element == 'H')
            He = (f.element == 'He')
            if e or H:
                neq = self.did_hion
            elif He:
                neq = self.params.get('do_helium', False)
            else:
                neq = False
            non_neq = (not neq)
            # logic:
            if mode == 'elem' and f.element is not None:
                result = 'elem'
            elif neq and (mode == 'best' or mode == 'neq'):
                result = 'neq'
            elif (not e) and ((mode == 'saha') or (mode == 'best' and non_neq)):
                result = 'saha'
            elif e:
                if (mode == 'table') or (mode == 'best' and non_neq):
                    result = 'table'
                elif (mode == 'QN_neq') or (mode == 'QN' and neq):
                    result = 'QN_neq'
                elif (mode == 'QN_table') or (mode == 'QN' and non_neq):
                    result = 'QN_table'
        if result == 'nan':
            self._handle_typevar_nan(errmsg=f"ntype, when fluid={self.fluid!r}, n_mode={self.n_mode!r}.")
        return xr.DataArray(result)

    NTYPE_TO_VAR = {**MhdMultifluidDensityLoader.NTYPE_TO_VAR,
        'neq': 'n_neq',
    }
    _NTYPE_TO_NONSIMPLE_DEPS = {**MhdMultifluidDensityLoader._NTYPE_TO_NONSIMPLE_DEPS,
        'QN_neq': ['ne_QN', 'ne_neq'],
    }
    # _NTYPE_TO_DEPS property is inherited from MhdMultifluidDensityLoader.

    @known_var(partition_across_dim=('fluid', 'ntype'), partition_deps='_NTYPE_TO_DEPS')
    def get_n(self, *, ntype):
        '''number density. Formula depends on fluid:
        if SINGLE_FLUID, n = (r / m), from SINGLE_FLUID r & m.
            default m is the abundance-weighted average particle mass; see help(self.get_m) for details.
        if Element, n = (r / m), where
            r is inferred from abundances combined with SINGLE_FLUID r, and
            m is element particle mass (fluid.m)
        if Specie, n depends on ntype, determined from self.n_mode (and self.ne_mode, if electron);
            see help(self.get_ntype) for details.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        # super() can handle the NTYPE_TO_VAR stuff, and most other stuff;
        #    here we only need to handle the new stuff in _NTYPE_TO_NONSIMPLE_DEPS.
        if ntype == 'QN_neq':
            return self('ne_QN', ne_mode='neq')
        else:
            return super().get_n(ntype=ntype)


    # # # NON-EQUILIBRIUM NUMBER DENSITY # # #

    @known_var(load_across_dims=['fluid'])
    def get_n_neq(self):
        '''number density of self.fluid specie(s); non-equilibrium values.
        Result depends on fluid:
            electron --> 'hionne'
            'H_I'    --> sum('n1', 'n2', 'n3', 'n4', 'n5')
            'H_II'   --> 'n6'
            'He_I'   --> 'nhe1'  (actually, exp('nhe1'); aux stores log values)
            'He_II'  --> 'nhe2'  (actually, exp('nhe2'); aux stores log values)
            'He_III' --> 'nhe3'  (actually, exp('nhe3'); aux stores log values)
            other --> crash with FormulaMissingError.
        the electron fluid is tested via fluid.is_electron(),
        while the other species are tested via name-matching to the names above.
        '''
        # load_across_dims for this one, instead of partition_across_dim,
        #   because here we don't expect multiple self.fluid with same formula,
        #   so there's basically no efficiency improvements from grouping.
        f = self.fluid
        if f.is_electron():
            return self('ne_neq')
        elif f == 'H_I':
            n1 = self('load_n1')
            n2 = self('load_n2')
            n3 = self('load_n3')
            n4 = self('load_n4')
            n5 = self('load_n5')
            result_cgs = n1 + n2 + n3 + n4 + n5
        elif f == 'H_II':
            result_cgs = self('load_n6')
        elif f == 'He_I':
            result_cgs = np.exp(self('load_nhe1'))
        elif f == 'He_II':
            result_cgs = np.exp(self('load_nhe2'))
        elif f == 'He_III':
            result_cgs = np.exp(self('load_nhe3'))
        else:
            raise FormulaMissingError(f'n_neq for fluid {f}.')
        result_cgs = self._upcast_if_max_n_requires_float64(result_cgs)  # <- maybe make float64.
        result = result_cgs * self.u('n', convert_from='cgs')  # <- convert to self.units system.
        return self.record_units(result)


    # # # NTYPE: SAHA # # #
    # inherited from MhdMultifluidDensityLoader


    # # # NTYPE: ELECTRON # # #

    _NE_TYPE_TO_DEPS = {**MhdMultifluidDensityLoader._NE_TYPE_TO_DEPS,
        'neq': ['ne_neq'],
        'QN_neq': ['ne_QN', 'ne_neq'],
    }

    @known_var(value_deps=[('ne_type', '_NE_TYPE_TO_DEPS')])
    def get_ne(self):
        '''electron number density. Result depends on self.ne_mode.
        See self.NE_MODE_OPTIONS for details.
        '''
        kind = self('ne_type').item()
        # 'neq' and 'QN_neq' are new in this subclass; all other kinds can be handled by super().
        if kind == 'neq':
            return self('ne_neq')
        elif kind == 'QN_neq':
            return self('ne_QN', ne_mode='neq')
        else:
            return super().get_ne()

    # get_ne_QN() is inherited from MhdMultifluidDensityLoader.

    @known_var(dims=['snap'], ignores_dims=['fluid'])
    def get_ne_neq(self):
        '''electron number density, from 'hionne' in aux.
        hionne in aux is stored in cgs units.
        '''
        result = super().get_ne_neq()  # see BifrostEosLoader
        return self._assign_electron_fluid_coord_if_unambiguous(result)



"""
File Purpose: os tools
e.g. for managing files & directories.
"""
import os
import re


''' --------------------- PlasmaCalcs path --------------------- '''

# CAUTION: hard-coded assumption that this file is in _PLASMACALCS_DIR/PlasmaCalcs/tools.
_HERE = os.path.dirname(__file__)
_PLASMACALCS_DIR = os.path.abspath(os.path.join(_HERE, '..', '..'))

def pc_path(*paths):
    '''PlasmaCalcs directory path, as abspath.
    returns os.path.join(path_to_PlasmaCalcs_directory, *paths).
    
    Examples:
        pc_path() # points to the PlasmaCalcs directory.
        pc_path('PlasmaCalcs/tools/os_tools') # points to the os_tools.py file.
        # points to the test_eppic_tinybox folder from tests/test_eppic.
        pc_path('tests', 'test_eppic', 'test_eppic_tinybox')
    '''
    return os.path.join(_PLASMACALCS_DIR, *paths)


''' --------------------- Finding files --------------------- '''

def find_files_re(pattern, dir=os.curdir, *, exclude=[]):
    '''find all files in this directory and all subdirectories which match the given pattern.

    pattern: str
        regular expression pattern to match filenames.
    exclude: str, or list of strs
        exclude any subdirectories whose name equals one of these strings or re.fullmatch one of these strings.
        E.g. exclude='*[.]io' will exclude all subdirectories whose name ends with '.io';
            exclude='parallel' will exclude all subdirectories whose name equals 'parallel'.
    
    returns list of abspaths to all files found.
    '''
    files = []
    pattern = re.compile(pattern)
    exclude = [exclude] if isinstance(exclude, str) else exclude
    exclude = [re.compile(str(excl)) for excl in exclude]
    for dirpath, dirnames, filenames in os.walk(dir, topdown=True):
        this_dir_name = os.path.basename(dirpath)
        if any((excl==this_dir_name or excl.fullmatch(this_dir_name)) for excl in exclude):
            dirnames[:] = []
            continue
        for filename in filenames:
            if pattern.fullmatch(filename):
                files.append(os.path.abspath(os.path.join(dirpath, filename)))
    return files


''' --------------------- Changing Directories --------------------- '''

class InDir():
    '''context manager for remembering directory.
    upon enter, cd to directory (default os.curdir, i.e. no change in directory)
    upon exit, original working directory will be restored.

    For function decorator, see tools.maintain_cwd.
    '''

    def __init__(self, directory=os.curdir):
        self.cwd = os.path.abspath(os.getcwd())
        self.directory = directory

    def __enter__(self):
        os.chdir(self.directory)

    def __exit__(self, exc_type, exc_value, traceback):
        os.chdir(self.cwd)


def with_dir(directory):
    '''returns a function decorator which:
    - changes current directory to <directory>.
    - runs function
    - changes back to original directory.
    '''
    def decorator(f):
        @functools.wraps(f)
        def f_but_enter_dir(*args, **kwargs):
            with InDir(directory):
                return f(*args, **kwargs)
        return f_but_enter_dir
    return decorator

# define a new function decorator, maintain_cwd, which maintains current directory:
maintain_cwd = with_dir(os.curdir)

maintain_directory = maintain_cwd  # alias
maintain_dir = maintain_cwd  # alias


''' --------------------- Path names --------------------- '''

def get_paths_with_common_start(common_start, dir=os.curdir, *, exclude_if=None):
    '''return list of abspaths to paths in dir which start with common_start.
    common_start: str
        if abspath, compare to abspaths from dir, else compare to relpaths from dir.
    exclude_if: None or callable
        if provided, exclude any paths for which exclude_if(path). (using path=abspath)
    '''
    listdir = sorted(os.listdir(dir))
    if os.path.isabs(common_start):
        paths = [os.path.abspath(os.path.join(dir, path)) for path in listdir]
        paths = [path for path in paths if path.startswith(common_start)]
    else:
        paths = [os.path.abspath(path) for path in listdir if path.startswith(common_start)]
    if exclude_if is not None:
        paths = [path for path in paths if not exclude_if(path)]
    return paths

def next_unique_name(desired_name, existing_names):
    '''return a name like desired_name but which is not in existing_names.
    If desired_name is in existing_names,
        Appends '_N' where N is smallest integer>0 that produces a unique name.
    '''
    if desired_name not in existing_names:
        return desired_name
    maybe_matches = [name for name in existing_names if name.startswith(desired_name)]
    i = 1
    while desired_name + f'_{i}' in maybe_matches:
        i += 1
    return desired_name + f'_{i}'


''' --------------------- File Memory Usage --------------------- '''

def nbytes_path(path=os.curdir):
    '''returns total size (in bytes) of all files implied by path.
    path: str
        file --> size of file
        directory --> total size of all files in directory and subdirectories, recursively.
    '''
    if os.path.isfile(path):
        return os.path.getsize(path)
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            if not os.path.islink(fp):  # only add if not a symbolic link
                total_size += os.path.getsize(fp)
    return total_size



"""
File Purpose: EppicDirectLoader
"""
import os
import math  # for lcm...

# required external modules
import numpy as np
import xarray as xr

# optional external modules
from ...tools import ImportFailed
try:
    import h5py
except ImportError as err:
    h5py = ImportFailed("h5py", err=err, locals=locals())
try:
    import openpmd_api as io
except ImportError as err:
    openpmd_api = ImportFailed("openpmd_api", err=err, locals=locals())

# internal modules
from ...dimensions import INPUT_SNAP
from ...errors import (
    LoadingNotImplementedError,
    InputError,
)
from ...quantities import DirectLoader, QuasineutralLoader
from ...tools import interprets_fractional_indexing
from ...defaults import DEFAULTS


''' --------------------- EppicDirectLoader --------------------- '''

class EppicDirectLoader(DirectLoader, QuasineutralLoader):
    '''manages loading data directly from eppic output files.

    input_deck: EppicInputDeck
        input deck. E.g. EppicInputDeck.from_file('eppic.i').
        Currently, requires that input_deck.filename is not None.
    read_mode: str
        how to read the files.
        Currently, must be 'h5' or 'h5_2'. See help(EppicDirectLoader.read_mode) for details.

    attributes of self (not available at __init__)
    '''
    # parent class ordering notes: no consistent mro() if QuasineutralLoader before DirectLoader.
    #  (related to DirectLoader being a SnapHaver, but not exactly sure why the error happens.
    #   but DirectLoader first here is fine; it doesn't overlap with QuasineutralLoader methods.)

    _h5_zfill = DEFAULTS.EPPIC.H5_SNAP_ZFILL
    _slice_maindims_in_load_direct = True   # [EFF] slice directly when reading h5 files, if using self.slices

    def __init__(self, input_deck, *, read_mode='h5', **kw_super):
        self.input_deck = input_deck
        self.read_mode = read_mode
        super().__init__(**kw_super)

    # # # READ MODE AND SNAPDIR # # #
    @property
    def read_mode(self):
        '''mode telling which files to read.
        Currently, must be 'h5' or 'h5_2'
        Maybe other modes will be added at some point.

        Options:
            'h5' --> read from .h5 files,
                    determine file format based on input_deck['hdf_output_arrays'].
            'h5_2' --> read from .h5 files,
                    assuming input_deck['hdf_output_arrays']==2.
        '''
        return self._read_mode
    @read_mode.setter
    def read_mode(self, value):
        if value != 'h5' and value != 'h5_2' and value != 'h5_3' and value != 'h5_4':
            raise ValueError(f'read_mode {value!r} not supported.')
        self._read_mode = value

    @property
    def full_read_mode(self):
        '''full read_mode, including hdf_output_arrays information.'''
        read_mode = self.read_mode
        if read_mode == 'h5':
            hdf_output_arrays = self.input_deck['hdf_output_arrays']
            return f'{read_mode}_{hdf_output_arrays}'
        else:
            return read_mode

    @property
    def snapdir(self):
        '''directory containing the snapshot files.
        If self.full_read_mode=='h5_2' or 'h5_3' or 'h5_4',
        this is '{self.dirname}/parallel'.
        Otherwise, this will crash with a NotImplementedError.
        '''
        if self.full_read_mode == 'h5_2' or self.full_read_mode == 'h5_3' or self.full_read_mode == 'h5_4':
            return os.path.join(self.input_deck.dirname, 'parallel')
        else:
            raise NotImplementedError(f'{type(self).__name__}.snapdir, when read_mode={self.full_read_mode!r}')

    def snap_filepath(self, snap=None):
        '''convert snap to full file path for this snap.

        snap: None, str, int, or Snap
            the snapshot number to load.
            if None, use self.snap.
        '''
        if self.full_read_mode == 'h5_2' or self.full_read_mode == 'h5_3' or self.full_read_mode == 'h5_4':
            return self._h5_2_filename(snap=snap)
        else:
            raise NotImplementedError(f'{type(self).__name__}.snap_filepath, when read_mode={self.full_read_mode!r}')


    # # # LOAD FROMFILE # # #
    def _var_for_load_fromfile(self, varname_internal):
        '''return var, suitably adjusted to pass into load_fromfile().
        Here:
            append self.component if self._loading_component
            append self.fluid.N if self._loading_fluid
        E.g. 'flux' turns into 'fluxz2' when loading 'z' component and fluid 2.
        '''
        result = varname_internal
        if getattr(self, '_loading_component', False):
            result = result + str(self.component)
        if getattr(self, '_loading_fluid', False):
            result = result + str(self.fluid.N)
        return result

    def load_fromfile(self, fromfile_var, *args, snap=None, **kw):
        '''return numpy array of fromfile_var, loaded directly from file.
        use self.full_read_mode to determine which file(s) / how to read them.

        fromfile_var: str
            the name of the variable to read, adjusted appropriately for loading fromfile.
            E.g., use 'fluxz2', not 'flux', to get flux for fluid 2 and component z.
            See also: self._var_for_load_fromfile().
        snap: None, str, int, or Snap
            the snapshot number to load. if None, use self.snap.

        Example:
            fromfile_var='fluxx1', snap=7, read_mode='h5_2'
                --> h5py.File('parallel/parallel000007.h5')['fluxx1'][:]
        '''
        snap = self._as_single_snap(snap)  # expect single snap when loading from file.
        if hasattr(snap, 'exists_for') and not snap.exists_for(self):
            result = xr.DataArray(self.snap_dim.NAN, attrs=dict(units='raw'))
            return self.assign_snap_coord(result)  # [TODO][EFF] assign coords when making array instead...
        if snap is INPUT_SNAP:
            raise LoadingNotImplementedError(f'var={fromfile_var!r}, when snap is INPUT_SNAP')
        full_read_mode = self.full_read_mode
        if full_read_mode == 'h5_2':
            return self._h5_2_load_fromfile(fromfile_var, *args, snap=snap, **kw)
        elif full_read_mode == 'h5_3' or full_read_mode == 'h5_4':
            return self._openpmd_load_fromfile(fromfile_var, *args, snap=snap, **kw)
        # reaching this line means the read mode is not recognized.
        raise LoadingNotImplementedError(f'unsupported full_read_mode: {full_read_mode!r}')

    def directly_loadable_vars(self, snap=None):
        '''return tuple of directly loadable variables.
        These are the variables that can be loaded directly from a file,
        using the current full_read_mode.

        snap: None, str, int, or Snap
            the snapshot number to load. if None, use self.snap.
        '''
        snap = self._as_single_snap(snap)  # expect single snap when loading from file.
        if hasattr(snap, 'exists_for') and not snap.exists_for(self):
            return np.nan
        full_read_mode = self.full_read_mode
        if full_read_mode == 'h5_2':
            result = list(self._h5_2_directly_loadable_vars(snap=snap))
        elif full_read_mode == 'h5_3' or full_read_mode == 'h5_4':
            result = list(self._openpmd_directly_loadable_vars(snap=snap))
        return result

    def wheref_loadable(self, var=None, snaps=None, *, set=False):
        '''returns snaps from self.snaps where 'n', 'flux', 'nvsqr', 'vdist', 'phi' values can be loaded,
        for all fluids currently in self.fluid (for non-phi vars).
        [EFF] method is IMPLICIT, based on var_out_subcycling details for each fluid.
            Does not actually check the files directly, because it's expensive to do that.

        [TODO] account for subsampling_info too.

        var: None, list of str, str from ('n', 'flux', 'nvsqr', 'vdist', 'phi'), or 'all'
            None --> return dict of answer for each of those vars
            'all' --> return SnapList where all vars are loadable.
            list --> return SnapList where all vars in list are loadable.
            str --> return SnapList for this var only.
        snaps: None, 'current', 'current_range', 'current_n', or SnapList
            the snaps to consider.
            None --> self.snaps.
            'current' --> self.snap
            'current_range' --> self.snaps.select_between(self.snap[0].t, self.snap[-1].t)
            'current_n' --> 'current_range', but evently downsample result size similar to current n snap.
                        (note: even downsampling isn't always possible; result size won't exactly match.
                         useful if you want "roughly N snaps". See example below.)
            E.g. if self.snap [10,20,30,...,200] (len=19), and self.snaps [0,...,300] (len=301),
                checking where_loadable(var) which is available every 4 snaps, result will be:
                None --> [0,4,8,...,300] (len=76)
                'current' --> [20,40,...,200] (len=10)  # because 10,30,50,... aren't divisible by 4
                'current_range' --> [12,16,20,...,200] (len=48)
                'current_n' --> [12,20,28,...,200] (len=24)
                                # because interprets_fractional_indexing(slice(None,None,1/19), L=48)
                                # gives slice(0, None, 2), which is then applied to current_range result.
        set: bool
            if True, also set self.snap = result.
        '''
        result = {}
        downsample = False
        if snaps is None:
            snaps = self.snaps
        elif isinstance(snaps, str):
            if snaps=='current':
                snaps = self.snap_list()
            elif snaps=='current_range' or snaps=='current_n':
                snaplist = self.snap_list()
                if snaps == 'current_n': downsample = len(snaplist)
                snaps = self.snaps.select_between(snaplist[0].t, snaplist[-1].t)
            else:
                raise InputError(f'got str snaps={snaps!r}, expected "current" or "current_range"')
        combine = False
        KNOWN = ('n', 'flux', 'nvsqr', 'vdist', 'phi')
        if var is None:
            vars_ = KNOWN
        elif isinstance(var, str):
            if var == 'all':
                combine = True
                vars_ = KNOWN
            else:
                vars_ = [var]
        else:  # list of vars
            combine = True
            vars_ = var
        unknown = [v for v in vars_ if v not in KNOWN]
        if len(unknown) > 0:
            raise InputError(f'unsupported var(s): {unknown!r}. Expected only vars from: {KNOWN!r}')
        nout = self.input_deck['nout']
        subcycles = {}
        FLUID_VARS = {
            'n': 'part_out_subcycle',
            'flux': 'flux_out_subcycle',
            'nvsqr': 'nvsqr_out_subcycle',
            'vdist': 'vdist_out_subcycle',
        }
        for v, subvar in FLUID_VARS.items():
            if v in vars_:
                subcycle = 1
                for fluid in self.iter_fluid():
                    subcycle = math.lcm(subcycle, fluid.get(subvar, 1))
                subcycles[v] = subcycle
        if 'phi' in vars_:
            subcycles['phi'] = self.input_deck.get('phi_out_subcycle', 1)
        if combine:  # combine the subcycles
            subcycle = 1
            for v, sc in subcycles.items():
                subcycle = math.lcm(subcycle, sc)
            var = 'COMBINED'
            subcycles = {var: subcycle}
        for v, subcycle in subcycles.items():
            if subcycle == 1:
                result[v] = snaps
            else:
                igood = [i for i, snap in enumerate(snaps) if int(snap.s) % (subcycle * nout) == 0]
                result[v] = snaps[igood]
        # downsample maybe
        if downsample:
            for v, rr in result.items():
                if len(rr) > downsample:  # actually downsample this one :)
                    slicer = interprets_fractional_indexing(slice(None,None,1/downsample), L=len(rr))
                    result[v] = rr[slicer]
        if isinstance(var, str):
            result = result[var]
        # output (maybe set self.snap first, too)
        if set:
            self.snap = result
        return result

    def where_loadable(self, var=None, snaps=None, *, set=False):
        '''returns snaps from self.snaps where 'n', 'flux', 'nvsqr', 'vdist', 'phi' values can be loaded,
        for ALL fluids in self.fluids.
        Equivalent:
            with self.using(fluid=None):
                return self.wheref_loadable(var=var, snaps=snaps)

        [EFF] method is IMPLICIT, based on var_out_subcycling details for each fluid.
            Does not actually check the files directly, because it's expensive to do that.

        [TODO] account for subsampling_info too.

        var: None, list of str, str from ('n', 'flux', 'nvsqr', 'vdist', 'phi'), or 'all'
            None --> return dict of answer for each of those vars
            'all' --> return SnapList where all vars are loadable.
            list --> return SnapList where all vars in list are loadable.
            str --> return SnapList for this var only.
        snaps: None, 'current', 'current_range', 'current_n', or SnapList
            the snaps to consider.
            None --> self.snaps.
            'current' --> self.snap
            'current_range' --> self.snaps.select_between(self.snap[0].t, self.snap[-1].t)
            'current_n' --> 'current_range', but evently downsample result size similar to current n snap.
                        (note: even downsampling isn't always possible; result size won't exactly match.
                         useful if you want "roughly N snaps". See example below.)
            E.g. if self.snap [10,20,30,...,200] (len=19), and self.snaps [0,...,300] (len=301),
                checking where_loadable(var) which is available every 4 snaps, result will be:
                None --> [0,4,8,...,300] (len=76)
                'current' --> [20,40,...,200] (len=10)  # because 10,30,50,... aren't divisible by 4
                'current_range' --> [12,16,20,...,200] (len=48)
                'current_n' --> [12,20,28,...,200] (len=24)
                                # because interprets_fractional_indexing(slice(None,None,1/19), L=48)
                                # gives slice(0, None, 2), which is then applied to current_range result.
        set: bool
            if True, also set self.snap = result.
        '''
        with self.using(fluid=None):
            return self.wheref_loadable(var=var, snaps=snaps, set=set)

    # # # H5_2 READ MODE # # #
    def _h5_2_load_fromfile(self, fromfile_var, *args__None, snap=None, **kw__None):
        '''return numpy array of var, loaded directly from file, using "h5_2" read_mode.
        This corresponds to h5 read mode with hdf_output_arrays=2.

        fromfile_var: str
            the name of the variable as stored in the snapshot.
        snap: None, str, int, or Snap
            the snapshot number to load.
            if None, use self.snap.

        Example:
            fromfile_var='fluxx1', snap=7
                --> h5py.File('parallel/parallel000007.h5')['fluxx1'][:]
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        filename = self._h5_2_filename(snap=snap)
        if not os.path.exists(filename):
            raise FileNotFoundError(f'{os.path.abspath(filename)!r}')
        with h5py.File(filename, 'r') as file:
            try:
                result = file[fromfile_var]
            except KeyError:
                errmsg = f'var={fromfile_var!r} not recognized (in file {os.path.abspath(filename)!r})'
                raise LoadingNotImplementedError(errmsg)
            if getattr(self, '_slice_maindims_in_load_direct', False):
                # [EFF] slice here instead of reading all data then slicing later.
                preslice = result
                result = self._slice_maindims_numpy(result, h5=True)
                return result[:] if (result is preslice) else result  # if didn't slice yet, use [:] to read h5 data.
            else:
                return result[:]

    def _h5_2_filename(self, *, snap=None):
        '''return name of file from which to load values, for read_mode='h5_2'.

        snap: None, str, int, or Snap
            the snapshot number to load.
            if None, use self.snap.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        snap = self._as_single_snap(snap)
        dir_ = self.snapdir  # probably {dirname}/parallel
        file_s = snap.file_s(self) if hasattr(snap, 'file_s') else str(snap)
        snap00N = file_s.zfill(self._h5_zfill)  # e.g. '000017' if _h5_zfill=6, snap=17.
        filename = os.path.join(dir_, f'parallel{snap00N}.h5')
        if (self.full_read_mode == 'h5_4'):
            filename = filename.replace('.h5', '.bp')
        return filename

    def _h5_2_directly_loadable_vars(self, *, snap=None):
        '''return tuple of directly loadable variables, for read_mode='h5_2'.'''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        filename = self._h5_2_filename(snap=snap)
        with h5py.File(filename, 'r') as file:
            return tuple(file.keys())

    # # # OPENPMD READ MODE # # #
    def _openpmd_directly_loadable_vars(self, *, snap=None):
        '''return tuple of directly loadable variables, for read_mode='openpmd'.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        filename = self._h5_2_filename(snap=snap)
        
        series = io.Series(filename, io.Access.read_only)
        snap = self._as_single_snap(snap)
        i = series.iterations[int(str(snap))]
        meshes = [m for m, n in i.meshes.items()]
        return tuple(meshes)
    
    def _openpmd_load_fromfile(self, eppic_var, *args__None, snap=None, **kw__None):
        '''return numpy array of var, loaded directly from file, using "openpmd" read_mode.
        This corresponds to openpmd read mode.

        eppic_var: str
            the name of the variable as stored in the snapshot.
        snap: None, str, int, or Snap
            the snapshot number to load.
            if None, use self.snap.

        Example:
            eppic_var='fluxx1', snap=7
                --> zarr.open('data/000007/fields/fluid_2/fluxx1')
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        
        snap = self._as_single_snap(snap)
        if eppic_var not in self._openpmd_directly_loadable_vars(snap=snap):
            errmsg = f'var={eppic_var!r} not recognized'
            raise LoadingNotImplementedError(errmsg)
        
        filename = self._h5_2_filename(snap=snap)
        series = io.Series(filename, io.Access.read_only)
        i = series.iterations[int(str(snap))]
        
        try:
            var = i.meshes[eppic_var]
        except KeyError:
            errmsg = f'var={eppic_var!r} not recognized (in file {os.path.abspath(filename)!r})'
            raise LoadingNotImplementedError(errmsg)

        if getattr(self, '_slice_maindims_in_load_direct', False):
            slices = self.slices
            slices = [slices.get(dim, slice(None)) for dim in self.maindims]
            result = var[tuple(slices)]
        else:
            result = var[:]

        series.flush()
        series.close()
        return np.array(result[:])


"""
File Purpose: select part of a list-like object based on values.
E.g. index of "closest" value, list of all values greater than some reference value.
"""
import numpy as np

from .docs_tools import format_docstring
from .iterables import is_iterable
from .oop_tools import Binding
from ..errors import InputError
from ..defaults import DEFAULTS

binding = Binding(locals())

_paramdocs = {
    'force_relative': '''force_relative: None, 'before', 'after', 'equal', or 'isclose'
        whether to also force a certain relationship between result and x:
        result <= x  (if 'before),
        result >= x  (if 'after'),
        result == x  (if 'equal'),
        np.isclose(result, x, atol=0)  (if 'isclose').''',
    }

@format_docstring(**_paramdocs)
def select_i_closest(listlike, x, force_relative=None):
    '''returns the index of the value closest to x. If x iterable, [select_i_closest(y) for y in x]
    raise ValueError if that is impossible (e.g. if force_relative='equal' and no value == x).
    {force_relative}
    '''
    if is_iterable(x):
        return [select_i_closest(listlike, x_, force_relative=force_relative) for x_ in x]
    ibest = None
    distbest = None
    if force_relative not in (None, 'before', 'after', 'equal', 'isclose'):
        errmsg = (f'force_relative={force_relative!r} invalid. '
                  'Expected None, "before", "after", "equal", or "isclose".')
        raise InputError(errmsg)
    if force_relative == 'equal':
        for i, x_ in enumerate(listlike):
            if x_ == x:
                ibest = i
                break
    elif force_relative == 'isclose':
        imatches = np.where(np.isclose(listlike, x, atol=0))[0]
        if len(imatches) == 1:
            ibest = imatches[0]
        elif len(imatches) >= 2:
            raise NotImplementedError(f'force_relative={force_relative!r} when multiple matches found.')
    else:
        for i, x_ in enumerate(listlike):
            dist = x_ - x   # negative if 'before', positive if 'after'
            if (force_relative == 'before' and dist >= 0) or \
               (force_relative == 'after' and dist <= 0):
                continue
            absdist = abs(dist)
            if (distbest is None) or (absdist < distbest):
                distbest = absdist
                ibest = i
    if ibest is None:
        raise ValueError(f'no compatible value found! (x={x}, force_relative={force_relative!r})')
    return ibest

@format_docstring(**_paramdocs)
def select_closest(listlike, x, force_relative=None):
    '''returns the value closest to x.
    If x iterable, [select_closest(y) for y in x]
    {force_relative}
    '''
    if is_iterable(x):
        return [select_closest(listlike, x_, force_relative=force_relative) for x_ in x]
    return listlike[select_i_closest(x, force_relative=force_relative)]

def select_i_before(listlike, x):
    '''returns the index of the value closest to x, but not larger than x.
    If x iterable, [select_i_before(y) for y in x]
    '''
    return select_i_closest(listlike, x, force_relative='before')

def select_before(listlike, x):
    '''returns the value closest to x, but not larger than x.
    If x iterable, [select_before(y) for y in x]
    '''
    return select_closest(listlike, x, force_relative='before')

def select_all_before(listlike, x):
    '''returns listlike, sliced until the value closest to x, but not larger than x.'''
    return listlike[:select_i_before(x)+1]

def select_i_after(listlike, x):
    '''returns the index of the value closest to x, but not smaller than x.
    If x iterable, [select_i_after(y) for y in x]
    '''
    return select_i_closest(listlike, x, force_relative='after')

def select_after(listlike, x):
    '''returns the value closest to x, but not smallerthan x.
    If x iterable, [select_after(y) for y in x]
    '''
    return select_closest(listlike, x, force_relative='after')

def select_all_after(listlike, x):
    '''returns listlike, sliced starting from the value closest to x, but not smaller than x.'''
    return listlike[select_i_after(x):]

def select_i_between(listlike, x1, x2):
    '''returns the list of indices for all values x1 <= value <= x2.'''
    return [i for i, value in enumerate(listlike) if x1 <= value <= x2]

def select_between(listlike, x1, x2):
    '''returns the list of values with x1 <= value <= x2.'''
    ii = select_i_between(x1, x2)
    try:
        return ll[ii]
    except TypeError:  # ll doesn't support fancy indexing
        return [ll[i] for i in ii]


class ArraySelectable():
    '''class which adds select_... methods to a listlike object.'''
    select_i_closest = select_i_closest
    select_closest = select_closest

    select_i_before = select_i_before
    select_before = select_before
    select_all_before = select_all_before

    select_i_after = select_i_after
    select_after = select_after
    select_all_after = select_all_after

    select_i_between = select_i_between
    select_between = select_between


class ArraySelectableChildHaver():
    '''class which provides select_... methods for a child object.

    indicate the attribute name of the child object via 'array_selectable_child' kwarg at subclass creation.
        E.g., class SnapList(..., array_selectable_child='x')
            --> self.select_i_closest(5) will return i such that self.x[i] is closest to 5.

    selections which return values (e.g. select_closest, but not select_i_closest)
        will always return the values from self, not the child.
        E.g. self.select_closest(5) returns self[i] such that self.x[i] is closest to 5.

    Implemented using __init_subclass__ to make pretty docstrings.
    '''
    def __init_subclass__(cls, *, array_selectable_child=None, **kw_super):
        '''called when subclassing ArraySelectableChildHaver; sets cls.select_... methods.
        array_selectable_child: None or str
            attribute name of child object which will be used for select_... methods.
            if None, no select_... methods will be added.

        the select_... methods are:
            select_i_closest, select_closest,
            select_i_before, select_before, select_all_before,
            select_i_after, select_after, select_all_after,
            select_i_between, select_between

        also sets self.array_selectable_child = array_selectable_child, for bookkeeping reasons.
        '''
        super().__init_subclass__(**kw_super)
        if array_selectable_child is None:
            return

        # else, array_selectable_child was provided!
        cls.array_selectable_child = array_selectable_child

        selector = format_docstring(child=array_selectable_child, **_paramdocs, sub_indent=DEFAULTS.TAB*3)

        with binding.to(cls):
            @binding
            @selector
            def select_i_closest(self, x, force_relative=None):
                '''returns the index of the value in self.{child} which is closest to x.
                If x is iterable, [self.select_i_closest(y) for y in x].
                {force_relative}
                '''
                child = getattr(self, self.array_selectable_child)
                return ArraySelectable.select_i_closest(child, x, force_relative=force_relative)

            @binding
            @selector
            def select_closest(self, x, force_relative=None):
                '''returns self[i] such that self.{child}[i] is closest to x.
                If x is iterable, [self.select_closest(y) for y in x].
                    If self supports fancy indexing, will use self[indices],
                    otherwise use type(self)[self[i] for i in indices].
                {force_relative}
                '''
                idx = self.select_i_closest(x, force_relative=force_relative)
                try:
                    return self[idx]
                except TypeError:  # self doesn't support fancy indexing
                    return type(self)(self[i] for i in idx)

            @binding
            @selector
            def select_i_before(self, x):
                '''returns the index of the value in self.{child} closest to x, but not larger than x.
                If x is iterable, [self.select_i_before(y) for y in x].
                '''
                return self.select_i_closest(x, force_relative='before')

            @binding
            @selector
            def select_before(self, x):
                '''returns self[i] such that self.{child}[i] is closest to x, but not larger than x.
                If x is iterable, [self.select_before(y) for y in x].
                    If self supports fancy indexing, will use self[indices],
                    otherwise use type(self)[self[i] for i in indices].
                '''
                return self.select_closest(x, force_relative='before')

            @binding
            @selector
            def select_all_before(self, x):
                '''returns self[:i] such that self.{child}[j] < x, for all j < i.'''
                return self[:self.select_i_before(x)+1]

            @binding
            @selector
            def select_i_after(self, x):
                '''returns the index of the value in self.{child} closest to x, but not smaller than x.
                If x is iterable, [self.select_i_after(y) for y in x].
                '''
                return self.select_i_closest(x, force_relative='after')

            @binding
            @selector
            def select_after(self, x):
                '''returns self[i] such that self.{child}[i] is closest to x, but not smaller than x.
                If x is iterable, [self.select_after(y) for y in x].
                    If self supports fancy indexing, will use self[indices],
                    otherwise use type(self)[self[i] for i in indices].
                '''
                return self.select_closest(x, force_relative='after')

            @binding
            @selector
            def select_all_after(self, x):
                '''returns self[i:] such that self.{child}[j] > x, for all j >= i.'''
                return self[self.select_i_after(x):]

            @binding
            @selector
            def select_i_between(self, x1, x2):
                '''returns the list of indices for all values in self.{child} with x1 <= value <= x2.'''
                child = getattr(self, self.array_selectable_child)
                return ArraySelectable.select_i_between(child, x1, x2)

            @binding
            @selector
            def select_between(self, x1, x2):
                '''returns the list of self[i] such that x1 <= self.{child}[i] <= x2.'''
                ii = self.select_i_between(x1, x2)
                try:
                    return self[ii]
                except TypeError:  # self doesn't support fancy indexing
                    return [self[i] for i in ii]



"""
File Purpose: BifrostDirectLoader
"""
import os

import numpy as np
import xarray as xr

from ...defaults import DEFAULTS
from ...errors import CacheNotApplicableError, SnapValueError, DimensionalityError
from ...tools import (
    simple_property,
    product,
)
from ...quantities import DirectLoader


''' --------------------- BifrostDirectLoader --------------------- '''

class BifrostDirectLoader(DirectLoader):
    '''manages loading data directly from bifrost output files.'''
    
    _slice_maindims_in_load_direct = True   # tell MainDimensionsHaver to let us handle slicing logic here.

    # # # PROPERTIES AFFECTING DIRECT LOADING # # #
    cls_behavior_attrs.register('stagger_direct', default=True)
    stagger_direct = simple_property('_stagger_direct', default=True,
            doc='''whether to stagger arrays to cell centers when loading directly from file.
            if all arrays are at cell centers, don't need to worry about stagger anymore.
            else, need to be sure to align arrays before doing calculations on them.''')

    cls_behavior_attrs.register('squeeze_direct', default=True)
    squeeze_direct = simple_property('_squeeze_direct', default=True,
            doc='''whether to squeeze arrays when loading directly from file.
            if True, remove dimensions of size 1, and alter self.maindims appropriately.''')

    def _maindims_post_squeeze(self):
        '''returns tuple of maindims remaining after squeezing if applicable. E.g. ('x', 'z').
        If not squeeze_direct or if 3D run, maindims will be ('x', 'y', 'z').
        Otherwise, remove any dims with size 1.
        '''
        if not self.squeeze_direct:
            return ('x', 'y', 'z')
        else:
            data_shape = (self.params['mx'], self.params['my'], self.params['mz'])
            return tuple(d for d, size in zip(('x', 'y', 'z'), data_shape) if size > 1)

    def _squeeze_direct_result__post_slicing(self, array):
        '''squeeze result of direct loading (maybe actually load_fromfile), after self.slices have been applied.
        only squeezes the dims for which self.data_array_shape shows size 1.
        returns array, possibly squeezed.
        '''
        to_squeeze = [d for d, size in zip(('x', 'y', 'z'), self.data_array_shape) if size == 1]
        if len(to_squeeze)>0:
            already_removed = self._slices_which_scalarize()  # list of dims which slicing removed.
            if array.ndim != 3 - len(already_removed):
                raise DimensionalityError('squeeze_direct cannot account for all dims.')
            array_dims = [x for x in ('x', 'y', 'z') if x not in already_removed]
            indexers = [0 if d in to_squeeze else slice(None) for d in array_dims]
            array = array[tuple(indexers)]   # e.g. equivalent: array[:, 0, :] if squeezing y only.
        return array

    def _slice_direct_numpy(self, array):
        '''self._slice_maindims_numpy(array), but using self.squeeze_direct=False.
        Can apply this directly to arrays loaded from memmaps.
        Cannot apply this to arrays after squeezing.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        with self.using(squeeze_direct=False):
            return self._slice_maindims_numpy(array)

    # # # DIRECT STAGGER INSTRUCTIONS # # #
    # {fromfile_var: ops to perform on var to align it to grid cell centers}
    _STAGGER_DIRECT_TO_CENTER_OPS = {
        # snap vars
        'r': None, 'e': None,  # None means no ops needed
        'px':'xup', 'py':'yup', 'pz':'zup',
        'bx':'xup', 'by':'yup', 'bz':'zup',
        # other vars: ??? (any other directly loadable vars that need staggering?)
        # (e.g.: 'efx': 'yup zup', 'efy': 'xup zup', 'efz': 'xup yup')
    }

    def _stagger_direct_to_center__then_slice(self, array, fromfile_var):
        '''stagger directly loaded data array (of fromfile_var values) to cell centers.
        Then slice, if self._slice_maindims_in_load_direct.

        Instructions for how to stagger to center depends on fromfile_var.
        if fromfile_var doesn't require any ops to center it, do nothing.

        array: numpy array
            the numpy array of data from file. (memmap is okay too.)
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        ops = self._STAGGER_DIRECT_TO_CENTER_OPS.get(fromfile_var, None)
        if ops is None:
            return self._slice_direct_numpy(array)
        # else, stagger array to cell centers, then slice.
        return self._stagger_direct_and_slice(array, ops)

    # # # FILE PATH INFO # # #
    @property
    def snapdir(self):
        '''directory containing the snapshot files.
        Here, gives self.dirname, because Bifrost outputs are stored at top-level of directory.
        '''
        return self.dirname

    def snap_filepath(self, snap=None):
        '''convert snap to full file path for this snap, i.e. to the snapname_NNN.idl file.'''
        snap = self._as_single_snap(snap)
        dir_ = self.snapdir
        file_s = snap.file_s(self) if hasattr(snap, 'file_s') else str(snap)
        filename = os.path.join(dir_, f'{self.snapname}_{file_s}.idl')
        return filename

    # # # DIRECT LOADING-RELATED INFO # # #
    def directly_loadable_vars(self, snap=None):
        '''return tuple of directly loadable variables for this snap.

        snap: None, str, int, Snap, or iterable indicating multiple snaps.
            the snapshot number to load. if None, use self.snap.
            if multiple snaps, return the result if all snaps have same result, else crash.
        '''
        kw = dict() if snap is None else dict(snap=snap)
        with self.using(**kw):
            multiple_snaps = self.snap_is_iterable()
            if multiple_snaps:
                snaps = self.snap
            else:
                snaps = [self._as_single_snap(snap)]
        result = None
        for snap in snaps:
            # result for this snap
            if hasattr(snap, 'exists_for') and not snap.exists_for(self):
                result_here = np.nan
            else:
                vpm = snap.var_paths_manager(self)
                result_here = vpm.vars
            # compare against known result
            if result is None:
                result = result_here
            elif result != result_here:
                errmsg = 'directly_loadable_vars differ between snaps. Retry with single snap instead.'
                raise SnapValueError(errmsg)
            # else, continue
        return result

    def var_paths_manager(self, snap=None):
        '''return VarPathsManager for this snap (or current snap if None provided)'''
        snap = self._as_single_snap(snap)
        return snap.var_paths_manager(self)

    @property
    def data_array_shape(self):
        '''shape of each data array. ('mx', 'my', 'mz') from self.params.
        Assumes shape is the same for all snaps; will crash if not.
        '''
        return (self.params['mx'], self.params['my'], self.params['mz'])

    @property
    def data_array_size(self):
        '''array size (number of elements) of each data array. (from data_array_shape)'''
        return product(self.data_array_shape)

    data_array_dtype = simple_property('_data_array_dtype', default=np.dtype('float32'),
            doc='''numpy dtype of each data array. default is np.dtype('float32').
            Used by np.memmap during self.load_fromfile.''')

    @property
    def data_array_nbytes(self):
        '''number of bytes in each data array. (from data_array_size and data_array_dtype)'''
        return self.data_array_size * self.data_array_dtype.itemsize

    data_array_order = simple_property('_data_array_order', default='F',
            doc='''ordering of data array axes: 'F' (Fortran) or 'C' (C order). Default 'F'.''')

    # # # LOAD FROMFILE # # #
    def _var_for_load_fromfile(self, varname_internal):
        '''return var, suitably adjusted to pass into load_fromfile().
        Here:
            append self.component if self._loading_component
        E.g. 'b' turns into 'bz' when loading 'z' component.
        '''
        result = varname_internal
        if getattr(self, '_loading_component', False):
            result = result + str(self.component)
        return result

    def load_fromfile(self, fromfile_var, *args__None, snap=None, **kw__None):
        '''return numpy array of fromfile_var, loaded directly from file.

        fromfile_var: str
            the name of the variable to read, adjusted appropriately for loading fromfile.
            E.g. use 'bz' not 'b', to get magnetic field z-component.
            See also: self._var_for_load_fromfile().
        snap: None, str, int, or Snap
            the snapshot number to load. if None, use self.snap.
        '''
        snap = self._as_single_snap(snap)  # expect single snap when loading from file.
        if hasattr(snap, 'exists_for') and not snap.exists_for(self):
            result = xr.DataArray(self.snap_dim.NAN, attrs=dict(units='raw'))
            return self.assign_snap_coord(result)  # [TODO][EFF] assign coords when making array instead...
        vpm = snap.var_paths_manager(self)
        filepath = vpm.var2path[fromfile_var]
        offset = vpm.var2index[fromfile_var] * self.data_array_nbytes
        result = np.memmap(filepath, offset=offset,
                           mode='r',  # read-only; never alters existing files!
                           shape=self.data_array_shape,
                           dtype=self.data_array_dtype,
                           order=self.data_array_order,
                           )
        result = np.asarray(result)  # convert to numpy array.
        if self.squeeze_direct:  # squeeze bookkeeping before applying any self.slices.
            assert result.shape == self.data_array_shape, "pre-squeeze shape check failed."
        if self.stagger_direct:
            result = self._stagger_direct_to_center__then_slice(result, fromfile_var)
        else:
            result = self._slice_direct_numpy(result)
        if self.squeeze_direct:
            result = self._squeeze_direct_result__post_slicing(result)
        return result



"""
File Purpose: calculating plasma drifts, e.g. hall, pederson.
"""
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from .plasma_parameters import PlasmaParametersLoader
from ..errors import FormulaMissingError
from ..plotting import PlotSettings
from ..tools import (
    simple_property,
    xarray_sum, xarray_sel,
)


class PlasmaDriftsLoader(PlasmaParametersLoader):
    '''plasma drifts.'''
    # # # GET DRIFTS # # #
    @known_var(deps=['skappa', 'E_un0_cross_B', 'mod_B'])
    def get_u_hall(self, *, _E=None, _B=None):
        '''Hall drift velocity. u_hall = (kappa**2 / (1 + kappa**2)) * (E_un0 x B) / |B|**2,
        where kappa is the magnetization parameter, kappa = gyrof / nusn.
        [EFF] for efficiency, can provide E and/or B, if already known.
        '''
        skappa = self('skappa')  # signed kappa; skappa<0 when q<0.
        E_cross_B = self('E_un0_cross_B', _E=_E, _B=_B)
        mod_B = self('mod_B', _B=_B)
        return ((skappa**2 / (1 + skappa**2)) / mod_B**2) * E_cross_B 

    @known_var(deps=['skappa', 'E_un0', 'mod_B'], aliases=['u_ped'])
    def get_u_pederson(self, *, _E=None, _B=None):
        '''Pederson drift velocity. u_pederson = (skappa / (1 + skappa**2)) * E_un0 / |B|,
        where skappa is the (signed) magnetization parameter, skappa = q * |B| / (m * nusn).
        [EFF] for efficiency, can provide E and/or B, if already known.
        '''
        skappa = self('skappa')  # signed kappa; skappa<0 when q<0.
        E = self('E_un0') if _E is None else _E
        mod_B = self('mod_B', _B=_B)
        return ((skappa / (1 + skappa**2)) / mod_B) * E 

    @known_var(deps=['skappa', 'E_un0_dot_B', 'B', 'mod_B'])
    def get_u_EdotB(self, *, _E=None, _B=None):
        '''EdotB drift velocity. u_EdotB = (skappa**3 / (1 + skappa**2)) * (E_un0 dot B) B / |B|^3
        (Commonly neglected, but comes from the same physical equation as hall & pederson drifts;
        from solving equilibrium momentum equation for u, when neglecting all derivatives.)
        [EFF] for efficiency, can provide E and/or B, if already known.
        '''
        skappa = self('skappa')
        E_dot_B = self('E_un0_dot_B', _E=_E, _B=_B)
        B = self('B') if _B is None else _B
        mod_B = self('mod_B', _B=_B)
        return ((skappa**3 / (1 + skappa**2)) * B / mod_B**3) * E_dot_B

    @known_var(deps=['u_hall', 'u_pederson', 'u_EdotB'], aliases=['u_eqcol', 'eqcol_u'])
    def get_u_drift(self):
        '''equilibrium velocity; solution to the momentum equation with collisions,
        assuming zero acceleration and zero spatial gradients.
        u_drift = u_hall + u_pederson + u_EdotB.
        '''
        # [EFF] calculate E & B first, to avoid recalculating them 3 times.
        #   will need all components of E & B, even if len(self.component)==1,
        #   because internally will calculate E cross B AND E dot B.
        with self.using(component=None):  # all components
            E = self('E_un0')
            B = self('B')
        result = self('u_hall', _E=E, _B=B) + self('u_pederson', _E=E, _B=B) + self('u_EdotB', _E=E, _B=B)
        if self.component != self.components:
            # if self.component is not None, then we need to select the right component(s).
            result = xarray_sel(result, component=self.component)
        return result
        # inefficient way: recalculates E & B, each time:
        # return self('u_hall') + self('u_pederson') + self('u_EdotB')

    @known_var(deps=['E_un0_perpmod_B', 'mod_B'])
    def get_EBspeed(self):
        '''speed determined from E_un0 and B: |E_un0 perp to B| / |B|.'''
        # [TODO] make a different variable for full |E_un0| (including parallel component)?
        # (the E dot B term of u_drift scales with skappa^3/(1+skappa^2) though,
        # which is not bounded between 0 and 1 like hall and pederson skappa-related coefficients...)
        return self('E_un0_perpmod_B') / self('mod_B')

    @known_var(deps=['J_perpmod_B', 'ne', 'abs_qe'])
    def get_JBspeed(self):
        '''speed determined from J perp to B, and ne: |J_perp to B| / (ne * |qe|).'''
        # [TODO] make a different variable for full |J| (including parallel component)?
        return self('J_perpmod_B') / (self('ne') * self('abs_qe'))

    # # # GET KAPPA & NUSN FROM VELOCITIES (assuming velocities are from drifts) # # #
    # note - example plotting code for these quantities can be found in eppic_moments.py

    @known_pattern(r'skappa_from_(means_|)(.*_)?momExB', ignores_dims=['component'],
                   deps=[{1: lambda groups: 'u' if groups[1] is None else groups[1][:-len('_')]},
                        'u_neutral', 'u_cross_B', 'mod_B', 'E_cross_B',])
    def get_skappa_from_momExB(self, var, *, _match=None):
        '''signed kappa (magnetization parameter) that statisfies momentum equation in the E x B direction.
        'skappa_from_{means_}{u_}momExB'
            E.g. 'skappa_from_means_momExB', 'skappa_from_momExB', 'skappa_from_means_moment1_momExB'
            {means_} = 'means_' or ''.
                if 'means_', take means of vars: 'u', 'u_neutral', 'mod_B', 'E_cross_B', 'u_cross_B'
            {u_} = '' or any other var then '_'.
                if provided, use this var instead of 'u' for velocity.  (Doesn't affect "u_neutral" though.)
                E.g. eppic calculator might use 'moment1' here, as in 'skappa_from_moment1_momExB'.

        Algebraic solution:
            momentum equation, rearranged using skappa = q * |B| / (m * nusn):
                0 = q (E + u x B) - m * nusn * (u - u_neutral)
                0 = skappa (E + u x B) - |B| (u - u_neutral)
            dotting with E x B:
                0 = skappa [(u x B) dot (E x B)] - |B| [(u - u_neutral) dot (E x B)]
                --> skappa = |B| [(u - u_neutral) dot (E x B)] / [(u x B) dot (E x B)]
        '''
        # Note, the denominator can be "simplified" further (but is not in this method) via:
        # (u x B) dot (E x B) = (u dot B) (E dot B) - (u dot E) |B|^2
        means, uvar = _match.groups()
        mean = 'mean_' if means else ''   # default ''
        ustr = uvar[:-len('_')] if uvar else 'u'
        with self.using(component=None):  # make sure to get all components of vectors.
            u = self(f'{mean}{ustr}')
            u_n = self(f'{mean}u_neutral')
            mod_B = self(f'{mean}mod_B')
            E_cross_B = self(f'{mean}E_cross_B')
            u_cross_B = self(f'{mean}{ustr}_cross_B')
            skappa = mod_B * self.dot(u - u_n, E_cross_B) / self.dot(u_cross_B, E_cross_B)
        return skappa

    @known_pattern(r'skappa_from_(means_|)(.*_)?momE', ignores_dims=['component'],
                   deps=[{1: lambda groups: 'u' if groups[1] is None else groups[1][:-len('_')]},
                        'u_neutral', 'E', 'mod_B', 'E_cross_B',])
    def get_skappa_from_momE(self, var, *, _match=None):
        '''signed kappa (magnetization parameter) that statisfies momentum equation in the E direction.
        'skappa_from_{means_}{u_}momE'
            E.g. 'skappa_from_means_momE', 'skappa_from_momE', 'skappa_from_means_moment1_momE'
            {means_} = 'means_' or ''.
                if 'means_', take means of vars: 'u', 'u_neutral', 'mod_E', 'mod_B', 'E', 'E_cross_B'
            {u_} = '' or any other var then '_'.
                if provided, use this var instead of 'u' for velocity.  (Doesn't affect "u_neutral" though.)
                E.g. eppic calculator might use 'moment1' here, as in 'skappa_from_moment1_momE'.

        Algebraic solution:
            momentum equation, rearranged using skappa = q * |B| / (m * nusn):
                0 = q (E + u x B) - m * nusn * (u - u_neutral)
                0 = skappa (E + u x B) - |B| (u - u_neutral)
            dotting with E:
                0 = skappa (|E|^2 + (u x B) dot E) - |B| (u - u_neutral) dot E    # note uxB.E == BxE.u == -ExB.u
                --> skappa = |B| (u - u_neutral) dot E / (|E|^2 - u dot (E x B))

        Note: results untrustworthy when kappa >> 1, since that involves dividing by a value close to 0.
        '''
        means, uvar = _match.groups()
        mean = 'mean_' if means else ''
        ustr = uvar[:-len('_')] if uvar else 'u'
        with self.using(component=None):  # make sure to get all components of vectors.
            u = self(f'{mean}{ustr}')
            u_n = self(f'{mean}u_neutral')
            E = self(f'{mean}E')
            mod_B = self(f'{mean}mod_B')
            E_cross_B = self(f'{mean}E_cross_B')
            mod_E_squared = self('mean_mod_E')**2 if means else self.dot(E,E)  # [EFF] use known E if not taking means.
            skappa = mod_B * self.dot(u - u_n, E) / (mod_E_squared - self.dot(u, E_cross_B))
        return skappa

    # [TODO] skappa from momB which gets skappa that satisfies momentum equation in the B direction.
    # (it is undetermined when E dot B == 0 == u dot B, so it's not useful if B perp to 2D simulation plane.)

    @known_pattern(r'skappa_from_(means_|)(.*_)?hall', ignores_dims=['component'],
                   deps=[{1: lambda groups: 'u' if groups[1] is None else groups[1][:-len('_')]},
                        'mod_B', 'E_cross_B', 'q'])
    def get_skappa_from_hall(self, var, *, _match=None):
        '''signed kappa (magnetization parameter) that statisfies u_hall = u, in the E x B direction.
        'skappa_from_{means_}{u_}hall'
            E.g. 'skappa_from_means_hall', 'skappa_from_hall', 'skappa_from_means_moment1_hall'
            {means_} = 'means_' or ''.
                if 'means_', take means of vars: 'u', 'u_neutral', 'mod_B', 'E_cross_B'
            {u_} = '' or any other var then '_'.
                if provided, use this var instead of 'u' for velocity.  (Doesn't affect "u_neutral" though.)
                E.g. eppic calculator might use 'moment1' here, as in 'skappa_from_moment1_hall'.

        Algebraic solution:
            formula for u_hall (from solving momentum equation for u in the E x B direction):
                u_hall = (kappa**2 / (1 + kappa**2)) * (E x B) / |B|**2
            solving for kappa**2, assuming u instead of u_hall, yields:
                (u dot (E x B)) == (kappa**2 / (1 + kappa**2)) * (|E x B|**2 / |B|**2)
                A + A * kappa**2 - kappa**2 == 0, where A = (u dot (E x B)) / (|E x B|**2 / |B|**2)
                kappa**2 = A / (1 - A)
                --> skappa = +- sqrt(A / (1 - A)),
            There are two solutions; return solution with the same sign as self('q') (i.e. fluid's charge)
        '''
        means, uvar = _match.groups()
        mean = 'mean_' if means else ''
        ustr = uvar[:-len('_')] if uvar else 'u'
        u_n = self(f'{mean}u_neutral')
        if np.any(u_n != 0):  # [TODO] account for nonzero u_neutral in this method.
            raise NotImplementedError(f'{var!r} when u_neutral != 0.')
        with self.using(component=None):  # make sure to get all components of vectors.
            u = self(f'{mean}{ustr}')
            mod_B = self(f'{mean}mod_B')
            E_cross_B = self(f'{mean}E_cross_B')
            A = self.dot(u, E_cross_B) / (self.dot(E_cross_B, E_cross_B) / mod_B**2)
            sign = np.sign(self('q'))
            skappa = sign * (A / (1 - A))**0.5
        return skappa

    @known_pattern(r'skappa_from_(means_|)(.*_)?pederson', ignores_dims=['component'],
                   deps=[{1: lambda groups: 'u' if groups[1] is None else groups[1][:-len('_')]},
                        'u_neutral', 'mod_B', 'E', 'q'])
    def get_skappa_from_pederson(self, var, *, _match=None):
        '''signed kappa (magnetization parameter) that statisfies u_pederson = u, in the E direction.
        'skappa_from_{means_}{u_}pederson'
            E.g. 'skappa_from_means_pederson', 'skappa_from_pederson', 'skappa_from_means_moment1_pederson'
            {means_} = 'means_' or ''.
                if 'means_', take means of vars: 'u', 'u_neutral', 'mod_B', 'E', 'mod_E',
            {u_} = '' or any other var then '_'.
                if provided, use this var instead of 'u' for velocity.  (Doesn't affect "u_neutral" though.)
                E.g. eppic calculator might use 'moment1' here, as in 'skappa_from_moment1_pederson'.

        Algebraic solution:
            formula for u_pederson (from solving momentum equation for u in the E direction):
                u_pederson = (skappa / (1 + skappa**2)) * E / |B|
            solving for skappa, assuming u instead of u_pederson, yields:
                (u dot E) == (skappa / (1 + skappa**2)) * (|E|**2 / |B|)
                A + A * skappa**2 - skappa == 0, where A = (u dot E) / (|E|**2 / |B|)
                skappa = (1 +- sqrt((-1)^2 - 4 * A * A)) / (2 * A),
            There are two solutions; the correct choice can be determined by using the momentum equation;
                the correct choice for the +- sign turns out to be: -sign(q) where q = self('q') == fluid's charge.

        Note: results untrustworthy when kappa >> 1, since that involves dividing by a value close to 0.
        '''
        means, uvar = _match.groups()
        mean = 'mean_' if means else ''
        ustr = uvar[:-len('_')] if uvar else 'u'
        u_n = self(f'{mean}u_neutral')
        if np.any(u_n != 0):  # [TODO] account for nonzero u_neutral in this method.
            raise NotImplementedError(f'{var!r} when u_neutral != 0.')
        with self.using(component=None):  # make sure to get all components of vectors.
            u = self(f'{mean}{ustr}')
            mod_B = self(f'{mean}mod_B')
            E = self(f'{mean}E')
            mod_E_squared = self('mean_mod_E')**2 if means else self.dot(E,E)  # [EFF] use known E if not taking means.
            A = self.dot(u, E) / (mod_E_squared / mod_B)
            qsign = np.sign(self('q'))
            skappa = (1 - qsign * (1 - 4 * A**2)**0.5) / (2 * A)
        return skappa

    @known_pattern(r'nusn_from_(means_|)(.*_)?(momExB|momE|hall|pederson)', ignores_dims=['component'],
                   deps=['sgyrof', {(0,1,2): 'skappa_from_{group0}{group1}{group2}'}])
    def get_nusn_from_drift(self, var, *, _match=None):
        '''nusn, calculated by assuming u satisfies momentum equation to zeroth order.
        There are various options for how to solve for nusn, as explained below (see {drift}).
        All solutions use nusn = sgyrof / skappa, where skappa is determined via
            skappa = self('skappa'+var[len('nusn'):]).
        E.g. 'nusn_from_means_momExB' --> use skappa = self('skappa_from_means_momExB').

        The description below helps explain the various options.

        'nusn_from_{means_}{u_}{drift}'
            E.g. 'nusn_from_means_momExB', 'nusn_from_hall', 'nusn_from_means_moment1_momB'
            {means_} = 'means_' or ''.
                if 'means_', take means of vars: 'sgyrof', any vars relevant to the chosen {drift} option.
            {u_} = '' or any other var then '_'.
                if provided, use this var instead of 'u' for velocity.  (Doesn't affect "u_neutral" though.)
            {drift} = 'momExB', 'momE', 'hall', or 'pederson'
                indicates how to solve for nusn. Use the similarly-named var when getting skappa.
                'momExB' --> get skappa from the momentum equation in the E x B direction.
                'momE' --> get skappa from the momentum equation in the E direction.
                'hall' --> get skappa from u_hall = u, in the E x B direction.
                'pederson' --> get skappa from u_pederson = u, in the E direction.
        '''
        means, uvar, drift = _match.groups()
        if uvar is None: uvar = ''
        kvar = f'skappa_from_{means}{uvar}{drift}'
        skappa = self(kvar)
        return self('sgyrof') / skappa

    def plot_check_nusn_from_drift(self, *, u='u', drift='momExB', cycle1=dict(ls=['-', '--', '-.', ':']),
                                   means=True, log=True, **kw_timelines):
        '''plots PlasmaCalcs.timelines() for comparing nusn to nusn inferred from drifts.
        This is meant to be used as a quick check. Use this code as an example if you need more low-level control.

        u: str or iterable of strs
            var to use for velocity. Might want something else, e.g. EppicCalculator might use u='moment1'
            iterable of strs --> get multiple.
        drift: str or iterable of strs
            tells the way to infer skappa, and thus nusn. Options: 'momExB', 'momE', 'hall', 'pederson'.
            iterable of strs --> get multiple.
        cycle1: dict of lists
            parameters to use for matplotlib plotting if getting multiple u or drift.
        means: bool
            whether to take means of lower-level vars while getting skappa.
            (e.g. use 'skappa_from_means_momExB' instead of 'skappa_from_momExB', if True.)
        log: bool
            whether to take log10 of the ratios (nusn_from_drift / nusn) before plotting.

        returns plt.gcf().
        '''
        mean = 'means_' if means else ''
        log10 = 'log10_' if log else ''
        if isinstance(u, str):
            u = [u]
        if isinstance(drift, str):
            drift = [drift]
        Nlines = len(u) * len(drift)
        Lcycle = None if len(cycle1)==0 else min(len(val) for val in cycle1.values())
        if (Lcycle is not None) and (Nlines > Lcycle):
            print(f'warning, too many timelines ({Nlines}) compared to style cycle length ({Lcycle}).')
        kw_timelines['ybounds'] = kw_timelines.pop('ybounds', PlotSettings.get_default('ybounds'))
        i = 0
        for uvar in u:
            for driftvar in drift:
                # calculation
                arr = self(f'{log10}(mean_(nusn_from_{mean}{uvar}_{driftvar}/nusn))')
                # plotting
                style = {k: v[i] for k,v in cycle1.items()}
                tls = arr.pc.timelines(label=f'({uvar}_{driftvar})', **style, **kw_timelines)
                # bookkeeping
                kw_timelines['ybounds'] = plt.ylim()  # ensure later timelines' ylims are big enough to show all timelines.
                i += 1
        # plot formatting
        plt.ylabel(f'{log10}(nusn_from_drift / nusn)')
        plt.title('check nusn from drift' if getattr(self, 'title', None) is None else self.title)

        return plt.gcf()

    # # # ELECTRIC FIELD FROM DRIFTS # # #
    @known_var(deps=['E0S1', 'E0S2', 'mod_B'])
    def get_eta0_J(self, *, _E0S1=None, _E0S2=None):
        '''E0_perp_B = eta0_J * J_perp_B + ..., where eta0_J = E0S1 * |B| / (E0S1^2 + E0S2^2). note: 
        see help(self.get_E0_un0_perpB) for more details.

        [EFF] for efficiency, can provide _E0S1 and/or _E0S2 if known.
        '''
        E0S1 = self('E0S1') if _E0S1 is None else _E0S1
        E0S2 = self('E0S2') if _E0S2 is None else _E0S2
        return E0S1 * self('mod_B') / (E0S1**2 + E0S2**2)

    @known_var(deps=['E0S1', 'E0S2', 'mod_B'])
    def get_eta0_hall(self, *, _E0S1=None, _E0S2=None):
        '''E0_perp_B = eta0_hall J x Bhat + ..., where eta0_hall = - E0S2 * |B| / (E0S1^2 + E0S2^2)
        see help(self.get_E0_un0_perpB) for more details.

        [EFF] for efficiency, can provide _E0S1 and/or _E0S2 if known.
        '''
        E0S1 = self('E0S1') if _E0S1 is None else _E0S1
        E0S2 = self('E0S2') if _E0S2 is None else _E0S2
        return -E0S2 * self('mod_B') / (E0S1**2 + E0S2**2)

    @known_var(deps=['q', 'n', 'skappa'], ignores_dims=['fluid'])
    def get_E0S1(self, *, _skappa=None, _n=None):
        '''E0S1 = sum_s (qs ns skappa_s / (1 + skappa_s^2)),
        summed across all charged fluids (from self.fluids).
        see help(self.get_E0_un0_perpB) for more details.

        [EFF] for efficiency, can provide _skappa and/or _n if known.
        '''
        with self.using(fluid=self.fluids.charged()):
            skappa = self('skappa') if _skappa is None else _skappa
            n = self('n') if _n is None else _n
            q = self('q')
        summand = q * n * skappa / (1 + skappa**2)
        return xarray_sum(summand, dim='fluid')

    @known_var(deps=['q', 'n', 'skappa'], ignores_dims=['fluid'])
    def get_E0S2(self, *, _skappa=None, _n=None):
        '''E0S2 = sum_s (qs ns skappa_s^2/ (1 + skappa_s^2)),
        summed across all charged fluids (from self.fluids).
        see help(self.get_E0_un0_perpB) for more details.

        [EFF] for efficiency, can provide _skappa and/or _n if known.
        '''
        with self.using(fluid=self.fluids.charged()):
            skappa = self('skappa') if _skappa is None else _skappa
            n = self('n') if _n is None else _n
            q = self('q')
        skappa2 = skappa**2
        summand = q * n * skappa2 / (1 + skappa2)
        return xarray_sum(summand, dim='fluid')

    @known_var(deps=['q', 'n', 'skappa'], ignores_dims=['fluid'])
    def get_E0S1_and_S2(self):
        '''dataset containing 'E0S1' and 'E0S2'. Equivalent to self(['E0S1', 'E0S2']),
        but more efficient (only computes skappa and n one time).
        E0S1 = sum_s (qs ns skappa_s / (1 + skappa_s^2))
        E0S2 = sum_s (qs ns skappa_s^2/ (1 + skappa_s^2))
        '''
        with self.using(fluid=self.fluids.charged()):
            skappa = self('skappa')
            n = self('n')
        return self(['E0S1', 'E0S2'], _skappa=skappa, _n=n)

    @known_var(deps=['eta0_J', 'J'])
    def get_E0_etaJ_perpB(self, *, _E0S1=None, _E0S2=None):
        '''E0_perp_B = E0_etaJ_perpB + ..., where E0_etaJ_perpB = eta0_J * J_perp_B.
        see help(self.get_E0_un0_perpB_fromJ) for more details.
        '''
        eta0_J = self('eta0_J', _E0S1=_E0S1, _E0S2=_E0S2)
        J = self('J_perp_B')
        return eta0_J * J

    @known_var(deps=['eta0_hall', 'J_cross_hat_B'])
    def get_E0_hall(self, *, _E0S1=None, _E0S2=None):
        '''E0_perp_B = E0_hall + ..., where E0_hall = eta0_hall * J x Bhat.
        see help(self.get_E0_un0_perpB) for more details.
        '''
        eta0_hall = self('eta0_hall', _E0S1=_E0S1, _E0S2=_E0S2)
        J_cross_Bhat = self('J_cross_hat_B')
        return eta0_hall * J_cross_Bhat

    @known_var(deps=['E0_etaJ_perpB', 'E0_hall'])
    def get_E0_un0_perpB(self):
        '''E0_un0_perpB = E0_etaJ_perpB + E0_hall
        == eta0_J * J_perp_B + eta0_hall * J x Bhat
        == (E0S1 |B| J_perp_B + E0S2 B cross J) / (E0S1^2 + E0S2^2), where
            E0S1 = sum_s (qs ns skappa_s / (1 + skappa_s^2)),
            E0S2 = sum_s (qs ns skappa_s^2/ (1 + skappa_s^2)),

        This is the electric field in the un=0 frame of reference, assuming:
            - zeroth order equilibrium velocities from multifluid equations,
            - including only collisions with neutrals,
            - considering only E (& J) perp to B; ignoring E0 (& J) parallel to B.
        '''
        E0S1_and_S2 = self('E0S1_and_S2')
        E0S1 = E0S1_and_S2['E0S1']
        E0S2 = E0S1_and_S2['E0S2']
        E0_etaJ = self('E0_etaJ_perpB', _E0S1=E0S1, _E0S2=E0S2)
        E0_hall = self('E0_hall', _E0S1=E0S1, _E0S2=E0S2)
        return E0_etaJ + E0_hall

    @known_var(aliases=['E0_un0_perpmagB'], deps=['mod_B', 'J_perpmod_B', 'E0S1', 'E0S2'])
    def get_E0_un0_perpmodB(self):
        '''E0_un0_perpmodB = |E0_un0_perpB|.
        Should be equivalent to self('mod_E0_un0_perpB'), aside from rounding errors.
        [EFF] the method here uses an algebraic simplification of the E0_un0_perpB formula:
            E0_un0_perpB == (E0S1 |B| J_perp_B + E0S2 B cross J) / (E0S1^2 + E0S2^2)
            --> (do some algebra, and utilizing J_perp_B dot B = 0) -->
            E0_un0_perpmodB = (|B| |J_perp_B|) / sqrt(E0S1^2 + E0S2^2)
        '''
        E0S1_and_S2 = self('E0S1_and_S2')
        E0S1 = E0S1_and_S2['E0S1']
        E0S2 = E0S1_and_S2['E0S2']
        mod_B = self('mod_B')
        mod_J = self('J_perpmod_B')
        return (mod_B * mod_J) / (E0S1**2 + E0S2**2)**0.5

    @known_var(aliases=['E0_un0_perpmagB_min'], deps=['mod_B', 'J_perpmod_B', 'nq'], ignores_dims=['fluid'])
    def get_E0_un0_perpmodB_min(self):
        '''E0_un0_perpmodB_min = minimum possible value of |E0_un0_perpB|, at each point.
            E0_un0_perpmodB_min = (1/sqrt(2)) * |B| |J_perp_B| / (ne |qe|)

        Regardless of kappa (& nusn) values for electrons and ions, will always have:
            E0_un0_perpmodB >= E0_un0_perpmodB_min.

        Logic which proves this fact (below, using "K" as shorthand for "skappa"):
            E0_un0_perpmodB has sqrt(E0S1^2 + E0S2^2) in the denominator;
            --> when sqrt(E0S1^2 + E0S2^2) is largest, E0_un0_perpmodB will be smallest.
            sqrt(E0S1^2 + E0S2^2) is largest when |E0S1| and |E0S2| are both largest.
            (1)  |E0S1| <= sum_s (|qs| ns |Ks| / (1 + Ks^2))
            (2)  |E0S2| <= sum_s (qs ns Ks^2 / (1 + Ks^2))
            For ANY real K, the relevant quantities are bounded by:
                (i)   0 < |K| / (1 + K^2) <= 1/2
                (ii)  0 < K^2 / (1 + K^2) < 1
                which can be readily shown using introductory calculus:
                    |K|/(1+K^2) has local extrema (derivative=0) only at K=1, where |K|/(1+K^2)=1/2;
                        it tends to 0 when K->0; and it tends to 0 when K->inf.
                    K^2/(1+K^2) has local extrema (derivative=0) only at K=0, where it equals 0;
                        it tends to 0 when K->0; and it tends to 1 when K->inf.
            Applying (i) to expression (1) above yields:
                |E0S1| <= sum_s |qs| ns * 1/2
                Utilizing quasineutrality (sum_s qs ns = 0) and qe < 0 and qi > 0, provides:
                    sum_s |qs| ns = ne |qe| + sum_i ni qi = 2 ne |qe|
                --> |E0S1| <= ne |qe|
            Meanwhile for expression (2), note that because qe < 0 and qi > 0,
                ne qe (Ke^2 / (1 + Ke^2)) has opposite the sign as sum_i ni qi Ki^2 / (1 + Ki^2),
                so the largest possible |E0S1| occurs when one of those two terms is as small as possible.
                Applying (ii) to each term, and quasineutrality to the second term, yields:
                    |ne qe (Ke^2 / (1 + Ke^2))| < ne |qe|
                    |sum_i ni qi Ki^2 / (1 + Ki^2)| < sum_i ni qi = ne |qe|
                --> |E0S2| < ne |qe|
            Combining yields:
                sqrt(E0S1^2 + E0S2^2) < sqrt(2) ne |qe|
            Thus, we always have:
                (|B| |J_perp_B|) / sqrt(E0S1^2 + E0S2^2) >= (|B| |J_perp_B|) / (sqrt(2) * ne |qe|)
                i.e.  E0_un0_perpmodB >= E0_un0_perpmodB_min.
        '''
        mod_B = self('mod_B')
        mod_J = self('J_perpmod_B')
        nqe = self('abs_nq', fluid=self.fluids.get_electron())
        if hasattr(nqe, 'drop_vars'):
            nqe = nqe.drop_vars('fluid', errors='ignore')
        return (mod_B * mod_J) / (np.sqrt(2) * nqe)

    @known_var(aliases=['E0_un0_perpmagB_simple'], deps=['mod_B', 'J_perpmod_B', 'nq'], ignores_dims=['fluid'])
    def get_E0_un0_perpmodB_simple(self):
        '''E0_un0_perpmodB_simple = simple estimate of |E0_un0_perpB|, at each point.
            E0_un0_perpmodB_simple = (|B| |J_perp_B|) / (ne |qe|)

        This is a simple estimate of |E0_un0_perpB|, accurate when kappae>>1 and kappai<<1.
        '''
        mod_B = self('mod_B')
        mod_J = self('J_perpmod_B')
        nqe = self('abs_nq', fluid=self.fluids.get_electron())
        if hasattr(nqe, 'drop_vars'):
            nqe = nqe.drop_vars('fluid', errors='ignore')
        return (mod_B * mod_J) / nqe

    # # # E_un0 mode # # #
    cls_behavior_attrs.register('E_un0_mode', default=None)
    _VALID_E_UN0_MODES = (None, 'un=u', 'un=0', 'E0_perpB', 'E0_perpmodB', 'E0_perpmodB_min')
    E_un0_mode = simple_property('_E_un0_mode', default=None, validate_from='_VALID_E_UN0_MODES',
        doc='''mode for calculating E_un0, the electric field in the neutral frame, where u_n=0.
        None, 'un=u', 'un=0', 'E0_perpB', 'E0_perpmodB', or 'E0_perpmodB_min'.
        None --> E_un0 = E + u_n x B. (i.e., E & u_n in "lab frame" then transform to u_n=0 frame.)
        'un=u' --> E_un0 = E + u x B. (i.e., assume un==u; transform to the u=0 frame.)
                    [EFF] if self.has_var('E_u0'), use E_un0 = self('E_u0') directly, without getting u;
                        assumes self('E_u0') provides "E without the u x B contribution".
                        If E_u0 not available, this mode requires 'u' to not vary with fluid.
        'un=0' --> E_un0 = E. (i.e., assume un==0 already; no need to shift frames.)
        'E0_perpB' --> E_un0 = E0_un0_perpB. (i.e., E perp to B in u_n=0 frame,
                    assuming zeroth order equilibrium velocities from multifluid equations,
                    including only collisions with neutrals,
                    and considering only E (& J) perp to B; ignoring E0 (& J) parallel to B.)
        'E0_perpmodB' --> |E_un0 perp to B| = E0_un0_perpmodB (== |E0_un0_perpB|). E_un0 = E0_un0_perpB.
                    [EFF] equivalent to using E0_perpB, except that, when getting |E_un0 perp to B|,
                    will use the more efficient formula for E0_un0_perpmodB.
        'E0_perpmodB_min' --> |E_un0 perp to B| = E0_un0_perpmodB_min. E_un0 = crash; cannot get E_un0 directly.
                    (i.e., minimum possible value of |E_un0 perp to B|, regardless of collision frequencies;
                    E0_un0_perpmodB_min = (1/sqrt(2)) * |B| |J_perp_B| / (ne |qe|).
                    See help(self.get_E0_un0_perpmodB_min) for details.)''')

    # some (but not all) of the E_un0_type options based on E_un0_mode.
    #   the types not included here are those with a more complicated dependency
    #   e.g. when mode == 'un=u' we need to check self.has_var('E_u0').
    _E_UN0_MODE_TO_TYPE = {
        None: 'E+unxB',
        'un=0': 'E',
        'E0_perpB': 'E0_perpB',
        'E0_perpmodB': 'E0_perpB',
        'E0_perpmodB_min': 'nan',  # can't get E_un0 directly; will crash.
    }

    @known_var
    def get_E_un0_type(self):
        '''string telling method that will be used to get E_un0. Based on self.E_un0_mode.
        possible results:
            'nan' <--> will crash. (e.g. this occurs if E_un0_mode = 'E0_perpmodB_min')
            'E+unxB' <--> self('E') + self('u_neutral_cross_B')
            'E_u0' <--> self('E_u0')
            'E+uxB' <--> self('E') + self('u_cross_B')
            'E' <--> self('E')
            'E0_perpB' <--> self('E0_un0_perpB')
        '''
        mode = self.E_un0_mode
        if mode in self._E_UN0_MODE_TO_TYPE:  # handles mode = None, 'un=0', 'E0_perpB', 'E0_perpmodB', 'E0_perpmodB_min'.
            result = self._E_UN0_MODE_TO_TYPE[mode]  # possible results = 'E+unxB', 'E', 'E0_perpB', 'nan'
        elif mode == 'un=u':
            if self.has_var('E_u0'):
                result = 'E_u0'
            else:
                result = 'E+uxB'
        else:
            result = 'nan'  # invalid E_un0 mode; should be prevented by E_un0 setter.
        if result == 'nan':
            self._handle_typevar_nan(errmsg=f"E_un0_type, when E_un0_mode=({mode!r}).")
        return xr.DataArray(result)

    _E_UN0_TYPE_TO_DEPS = {'E+unxB': ['E', 'u_neutral_cross_B'],
                           'E_u0': ['E_u0'],
                           'E+uxB': ['E', 'u_cross_B'],
                           'E': ['E'],
                           'E0_perpB': ['E0_un0_perpB']}

    @known_var(value_deps=[('E_un0_type', '_E_UN0_TYPE_TO_DEPS')])
    def get_E_un0(self):
        '''E_un0 = electric field in the neutral frame, where u_n=0.
        Result depends on self.E_un0_mode; see help(type(self).E_un0_mode) for details.
        '''
        kind = self('E_un0_type').item()
        if kind == 'nan':
            if self.E_un0_mode == 'E0_perpmodB_min':
                errmsg = ('E_un0 cannot be directly calculated when E_un0_mode="E0_perpmodB_min";\n'
                            'use a different E_un0_mode or get E_un0_perpmod_B instead.')
                raise FormulaMissingError(errmsg)
            else:
                raise LoadingNotImplementedError('unknown error when getting self.E_un0')
        elif kind == 'E+unxB':
            # [EFF] if un==0 just return E, without adding un x B.
            un = self('u_neutral', component=None)
            result = self('E')
            if np.any(un != 0):
                result = result + self('u_neutral_cross_B', _u_neutral=un)
        elif kind == 'E_u0':
            return self('E_u0')
        elif kind == 'E+uxB':
            # [EFF] if u==0 just return E, without adding u x B.
            u = self('u', component=None)
            result = self('E')
            if np.any(u != 0):
                result = result + self('u_cross_B', _u=u)
        elif kind == 'E':
            result = self('E')
        else:
            assert kind == 'E0_perpB', f'invalid E_un0_type ({kind!r}).'
            result = self('E0_un0_perpB')
        return result

    _E_UN0_MODE_TO_PERPMOD_B_DEPS = {'E0_perpmodB': 'E0_un0_perpmodB',
                                     'E0_perpmodB_min': 'E0_un0_perpmodB_min',
                                     '__default__': ['E_un0', 'B']}

    @known_var(aliases=['E_un0_perpmag_B'],
               attr_deps=[('E_un0_mode', '_E_UN0_MODE_TO_PERPMOD_B_DEPS')])
    def get_E_un0_perpmod_B(self, **kw_perpmod):
        '''E_un0_perpmod_B == |E_un0 perp to B| == magnitude of E, perp to B, in u_neutral=0 frame.
        This is usually equivalent to using self.magnitude(self.take_perp_to(B, E_un0)),
            but if E_un0_mode = 'E0_perpmodB' will use more efficient method (self('E0_un0_perpmodB')),
            and if E_un0_mode = 'E0_perpmodB_min' will use different method (self('E0_un0_perpmodB_min')).

        kwargs are passed to self.get_perpmod, if applicable.
        '''
        mode = self.E_un0_mode
        if mode == 'E0_perpmodB':
            return self('E0_un0_perpmodB')
        elif mode == 'E0_perpmodB_min':
            return self('E0_un0_perpmodB_min')
        else:  # use the logic directly from the perpmod pattern (see self.get_perpmod, for details)
            # [TODO] encapsulate this somehow? It's repeated code from QuantityLoader.__call__.
            try:
                matched = self.match_var('E_un0_perpmod_B', check='KNOWN_PATTERNS')
            except FormulaMissingError as err0:  # (direct)   # no match in KNOWN_VARS & KNOWN_PATTERNS.
                # >> actually get value from load_direct. Crash if that fails <<
                try:
                    return self.load_direct('E_un0_perpmod_B')
                except Exception as err1:
                    raise err1 from err0   # traceback will include info from err0 and err1.
            else:
                return matched.load_value(self, **kw_perpmod)



"""
File Purpose: values of constants (e.g. kB)

(Be careful not to use ambiguous letters here which might be used for other values,
    e.g. might want to avoid using 'e' to mean 2.728...
    since 'e' might be used for a variable like energy density.
    In those cases, can use a more descriptive name, e.g. 'econst' or 'eulerconst')
"""
import numpy as np

from .quantity_loader import QuantityLoader

class ConstantsLoader(QuantityLoader):
    '''load values of constants (numbers which depend on, at most, self.units)
    [EFF] note, it's more efficient to get these numbers directly inside functions,
        instead of calling self an extra time to load the constant.
        E.g. self.u('kB') instead of self('kB').
    However, it's sometimes very convenient to be able to access these values
        directly from the call architecture. E.g. when testing different variables
        and making plots based on them, such as self('m*(vsqr-u**2)/kB')
    '''
    @known_var
    def get_pi(self):
        '''pi. numpy.pi. 3.1415...'''
        return np.pi

    @known_var
    def get_kB(self):
        '''boltzmann constant, in [self.units] units. Equivalent to self.u('kB')'''
        return self.u('kB')

    @known_var
    def get_amu(self):
        '''1 atomic mass unit, in [self.units] units. Equivalent to self.u('amu')'''
        return self.u('amu')

    @known_var(deps=['m'])
    def get_m_amu(self):
        '''mass in atomic mass units. Equivalent to self('m/amu').'''
        return self('m/amu')

    @known_var
    def get_eps0(self):
        '''vacuum permittivity, in [self.units] units. Equivalent to self.u('eps0')'''
        return self.u('eps0')

    @known_var
    def get_abs_qe(self):
        '''|electron charge|, in [self.units] units. Equivalent to self.u('qe').
        abs to avoid any possible ambiguity, since electron charge itself is negative.
        '''
        return self.u('qe')
